# Copyright: (c) 2022, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import struct
import typing as t


def F(x: int, y: int, z: int) -> int:
    return (x & y) | (~x & z)


def G(x: int, y: int, z: int) -> int:
    return (x & y) | (x & z) | (y & z)


def H(x: int, y: int, z: int) -> int:
    return x ^ y ^ z


def ROTL(x: int, n: int) -> int:
    return ((x << n) & 0xFFFFFFFF) | (x >> (32 - n))


def FF(a: int, b: int, c: int, d: int, x: int, s: int) -> int:
    return ROTL((a + F(b, c, d) + x) & 0xFFFFFFFF, s)


def GG(a: int, b: int, c: int, d: int, x: int, s: int) -> int:
    return ROTL((a + G(b, c, d) + x + 0x5A827999) & 0xFFFFFFFF, s)


def HH(a: int, b: int, c: int, d: int, x: int, s: int) -> int:
    return ROTL((a + H(b, c, d) + x + 0x6ED9EBA1) & 0xFFFFFFFF, s)


def md4(data: bytes) -> bytes:
    """Python implementation of md4 hashing.

    This is a pure Python implementation of the `MD4 Hashing Algorithm`_.
    Recent distributions of Linux ship with OpenSSL 3.x which has disabled the
    md4 engine so for backwards compatibility this must be manually implemented
    here to ensure NTLM continues to work.

    .. _MD4 Hasing Algorithm:
        https://datatracker.ietf.org/doc/html/rfc1320
    """
    data_block = bytearray(data)
    orig_length = len(data)
    b_orig_length = ((orig_length * 8) & 0xFFFFFFFFFFFFFFFF).to_bytes(8, "little")

    # Need to pad the last block to 64 bytes, the first bit is set to 1 and the
    # last 8 bytes are the length of the input data.
    data_block += b"".join(
        [
            b"\x80",
            bytes((55 - orig_length % 64) % 64),
            b_orig_length,
        ]
    )

    A = 0x67452301
    B = 0xEFCDAB89
    C = 0x98BADCFE
    D = 0x10325476

    while data_block:
        AA, BB, CC, DD = A, B, C, D
        X = struct.unpack("<16I", data_block[:64])
        data_block = data_block[64:]

        # Round 1
        AA = FF(AA, BB, CC, DD, X[0], 3)
        DD = FF(DD, AA, BB, CC, X[1], 7)
        CC = FF(CC, DD, AA, BB, X[2], 11)
        BB = FF(BB, CC, DD, AA, X[3], 19)

        AA = FF(AA, BB, CC, DD, X[4], 3)
        DD = FF(DD, AA, BB, CC, X[5], 7)
        CC = FF(CC, DD, AA, BB, X[6], 11)
        BB = FF(BB, CC, DD, AA, X[7], 19)

        AA = FF(AA, BB, CC, DD, X[8], 3)
        DD = FF(DD, AA, BB, CC, X[9], 7)
        CC = FF(CC, DD, AA, BB, X[10], 11)
        BB = FF(BB, CC, DD, AA, X[11], 19)

        AA = FF(AA, BB, CC, DD, X[12], 3)
        DD = FF(DD, AA, BB, CC, X[13], 7)
        CC = FF(CC, DD, AA, BB, X[14], 11)
        BB = FF(BB, CC, DD, AA, X[15], 19)

        # Round 2
        AA = GG(AA, BB, CC, DD, X[0], 3)
        DD = GG(DD, AA, BB, CC, X[4], 5)
        CC = GG(CC, DD, AA, BB, X[8], 9)
        BB = GG(BB, CC, DD, AA, X[12], 13)

        AA = GG(AA, BB, CC, DD, X[1], 3)
        DD = GG(DD, AA, BB, CC, X[5], 5)
        CC = GG(CC, DD, AA, BB, X[9], 9)
        BB = GG(BB, CC, DD, AA, X[13], 13)

        AA = GG(AA, BB, CC, DD, X[2], 3)
        DD = GG(DD, AA, BB, CC, X[6], 5)
        CC = GG(CC, DD, AA, BB, X[10], 9)
        BB = GG(BB, CC, DD, AA, X[14], 13)

        AA = GG(AA, BB, CC, DD, X[3], 3)
        DD = GG(DD, AA, BB, CC, X[7], 5)
        CC = GG(CC, DD, AA, BB, X[11], 9)
        BB = GG(BB, CC, DD, AA, X[15], 13)

        # Round 3
        AA = HH(AA, BB, CC, DD, X[0], 3)
        DD = HH(DD, AA, BB, CC, X[8], 9)
        CC = HH(CC, DD, AA, BB, X[4], 11)
        BB = HH(BB, CC, DD, AA, X[12], 15)

        AA = HH(AA, BB, CC, DD, X[2], 3)
        DD = HH(DD, AA, BB, CC, X[10], 9)
        CC = HH(CC, DD, AA, BB, X[6], 11)
        BB = HH(BB, CC, DD, AA, X[14], 15)

        AA = HH(AA, BB, CC, DD, X[1], 3)
        DD = HH(DD, AA, BB, CC, X[9], 9)
        CC = HH(CC, DD, AA, BB, X[5], 11)
        BB = HH(BB, CC, DD, AA, X[13], 15)

        AA = HH(AA, BB, CC, DD, X[3], 3)
        DD = HH(DD, AA, BB, CC, X[11], 9)
        CC = HH(CC, DD, AA, BB, X[7], 11)
        BB = HH(BB, CC, DD, AA, X[15], 15)

        A = (A + AA) & 0xFFFFFFFF
        B = (B + BB) & 0xFFFFFFFF
        C = (C + CC) & 0xFFFFFFFF
        D = (D + DD) & 0xFFFFFFFF

    return struct.pack("<IIII", A, B, C, D)



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)


from spnego._sspi_raw.sspi import (
    ClientContextAttr,
    ClientContextReq,
    Credential,
    CredentialUse,
    SecBuffer,
    SecBufferDesc,
    SecBufferType,
    SecPkgAttr,
    SecPkgAttrSizes,
    SecPkgInfo,
    SecStatus,
    SecurityContext,
    ServerContextAttr,
    ServerContextReq,
    SSPIQoP,
    TargetDataRep,
    WinNTAuthIdentity,
    accept_security_context,
    acquire_credentials_handle,
    decrypt_message,
    encrypt_message,
    initialize_security_context,
    make_signature,
    query_context_attributes,
    verify_signature,
)

__all__ = [
    "ClientContextAttr",
    "ClientContextReq",
    "Credential",
    "CredentialUse",
    "SecBuffer",
    "SecBufferDesc",
    "SecBufferType",
    "SecPkgAttr",
    "SecPkgAttrSizes",
    "SecPkgInfo",
    "SecStatus",
    "SecurityContext",
    "ServerContextAttr",
    "ServerContextReq",
    "SSPIQoP",
    "TargetDataRep",
    "WinNTAuthIdentity",
    "accept_security_context",
    "acquire_credentials_handle",
    "decrypt_message",
    "encrypt_message",
    "initialize_security_context",
    "make_signature",
    "query_context_attributes",
    "verify_signature",
]



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from spnego._context import ContextProxy, ContextReq
from spnego._credential import Credential, NTLMHash, unify_credentials
from spnego._credssp import CredSSPProxy
from spnego._gss import GSSAPIProxy
from spnego._negotiate import NegotiateProxy
from spnego._ntlm import NTLMProxy
from spnego._sspi import SSPIProxy
from spnego.channel_bindings import GssChannelBindings
from spnego.exceptions import NegotiateOptions


def _new_context(
    username: typing.Optional[typing.Union[str, Credential, typing.List[Credential]]],
    password: typing.Optional[str],
    hostname: str,
    service: str,
    channel_bindings: typing.Optional[GssChannelBindings],
    context_req: ContextReq,
    protocol: str,
    options: NegotiateOptions,
    usage: str,
    **kwargs: typing.Any,
) -> ContextProxy:
    proto = protocol.lower()
    credentials = unify_credentials(username, password)
    sspi_protocols = SSPIProxy.available_protocols(options=options)
    gssapi_protocols = GSSAPIProxy.available_protocols(options=options)

    # Unless otherwise specified, we always favour the platform implementations (SSPI/GSSAPI) if they are available.
    # Otherwise fallback to the Python implementations (NegotiateProxy/NTLMProxy).
    use_flags = (
        NegotiateOptions.use_sspi
        | NegotiateOptions.use_gssapi
        | NegotiateOptions.use_negotiate
        | NegotiateOptions.use_ntlm
    )
    use_specified = options & use_flags != 0

    # Filter protocols that aren't compatible with the credentials provided.
    sspi_remove = set()
    for cred in credentials:
        # NTLMHash cannot be used with SSPI. The code falls back to using NTLMProxy if this credential is present.
        if isinstance(cred, NTLMHash):
            sspi_remove.add("negotiate")
            sspi_remove.add("ntlm")

    if sspi_remove:
        for protocol in sspi_remove:
            if protocol in sspi_protocols:
                sspi_protocols.remove(protocol)

    proxy: typing.Type[typing.Union[CredSSPProxy, NTLMProxy, SSPIProxy, GSSAPIProxy, NegotiateProxy]]

    # If the protocol is CredSSP then we can only use CredSSPProxy. The use_flags still control what underlying
    # Negotiate auth is used in the CredSSP authentication process.
    if proto == "credssp":
        proxy = CredSSPProxy

    elif options & NegotiateOptions.use_sspi or (not use_specified and proto in sspi_protocols):
        proxy = SSPIProxy

    elif options & NegotiateOptions.use_gssapi or (
        not use_specified and (proto == "kerberos" or proto in gssapi_protocols)
    ):
        proxy = GSSAPIProxy

    elif options & NegotiateOptions.use_negotiate or (not use_specified and proto == "negotiate"):
        # If GSSAPI does not offer full negotiate support, use our own wrapper.
        proxy = NegotiateProxy

    elif options & NegotiateOptions.use_ntlm or (not use_specified and proto == "ntlm"):
        # Finally if GSSAPI does not support ntlm, use our own wrapper.
        proto = "ntlm" if proto == "negotiate" else proto
        proxy = NTLMProxy

    else:
        raise ValueError("Invalid protocol specified '%s', must be kerberos, negotiate, or ntlm" % protocol)

    return proxy(
        username=credentials,
        password=None,
        hostname=hostname,
        service=service,
        channel_bindings=channel_bindings,
        context_req=context_req,
        usage=usage,
        protocol=proto,
        options=options,
        **kwargs,
    )


def client(
    username: typing.Optional[typing.Union[str, Credential, typing.List[Credential]]] = None,
    password: typing.Optional[str] = None,
    hostname: str = "unspecified",
    service: str = "host",
    channel_bindings: typing.Optional[GssChannelBindings] = None,
    context_req: ContextReq = ContextReq.default,
    protocol: str = "negotiate",
    options: NegotiateOptions = NegotiateOptions.none,
    **kwargs: typing.Any,
) -> ContextProxy:
    """Create a client context to be used for authentication.

    Credentials can be provides in 3 different ways:

        * omitted entirely
        * `username` set to a string with an optional password
        * `username` set to a single or list of :class:`Credential` objects

    If the credential is omitted the authentication protocol will use a
    protocol specific mechanism to find the credential in some cache. If no
    cache or credentials are found an exception is raised.

    If `username` is a string but `password` is not defined then the provider
    specific cache is used but for the principal specified. If `password` is
    specified then a new credential is retrieved using the username/password
    combiniation. Using both `username` and `password` is the same as using
    :class:`Password` as a credential.

    If `username` is a, or list of, `Credential` object the object(s) are
    passed into the authentication implementation for use. The behaviour of
    these credentials are defined by the authentication protocol
    implementation that uses the credential(s). Currently the :class:`Passowrd`
    credential is supported by all authentication protocols and is the same
    as specifying the username/password as separate arugments.

    The ``negotiate`` protocol can be given a list of credentials for each of
    the sub auth protocols it negotiates to use. For example to have Negotiate
    use a Kerberos CCache for Kerberos and an NT/LM hash value for NTLM the
    following can be specified::

        import spnego

        credential = [
            spnego.KerberosCCache(ccache="FILE:/tmp/my-ccache"),
            spnego.NTLMHash(
                username="user",
                nt_hash="8ADB9B997580D69E69CAA2BBB68F4697",
            )
        ]
        client = spnego.auth(credential, hostname="server")

    In the example above, the negotiate protocol will attempt use the CCache
    for Kerberos auth and fallback to the NT/LM hash for NTLM auth if
    Kerberos fails. If a credential for a specific sub protocol is not
    available then Negotiate will move onto the next available one.

    Args:
        username: The username/credential(s) to authenticate with. Certain providers can use a cache if omitted.
        password: The password to authenticate with. Should only be specified when username is a string.
        hostname: The principal part of the SPN. This is required for Kerberos auth to build the SPN.
        service: The service part of the SPN. This is required for Kerberos auth to build the SPN.
        channel_bindings: The optional :class:`spnego.channel_bindings.GssChannelBindings` for the context.
        context_req: The :class:`spnego.ContextReq` flags to use when setting up the context.
        protocol: The protocol to authenticate with, can be `ntlm`, `kerberos`, `negotiate`, or `credssp`.
        options: The :class:`spnego.NegotiateOptions` that define pyspnego specific options to control the negotiation.
        kwargs: Optional arguments to pass through to the authentiction context.

    Returns:
        ContextProxy: The context proxy for a client.
    """
    return _new_context(
        username, password, hostname, service, channel_bindings, context_req, protocol, options, "initiate", **kwargs
    )


def server(
    hostname: str = "unspecified",
    service: str = "host",
    channel_bindings: typing.Optional[GssChannelBindings] = None,
    context_req: ContextReq = ContextReq.default,
    protocol: str = "negotiate",
    options: NegotiateOptions = NegotiateOptions.none,
    **kwargs: typing.Any,
) -> ContextProxy:
    """Create a server context to be used for authentication.

    Args:
        hostname: The principal part of the SPN. This is required for Kerberos auth to build the SPN.
        service: The service part of the SPN. This is required for Kerberos auth to build the SPN.
        channel_bindings: The optional :class:`spnego.channel_bindings.GssChannelBindings` for the context.
        context_req: The :class:`spnego.ContextReq` flags to use when setting up the context.
        protocol: The protocol to authenticate with, can be `ntlm`, `kerberos`, `negotiate`, or `credssp`.
        options: The :class:`spnego.NegotiateOptions` that define pyspnego specific options to control the negotiation.
        kwargs: Optional arguments to pass through to the authentiction context.

    Returns:
        ContextProxy: The context proxy for a client.
    """
    return _new_context(
        None, None, hostname, service, channel_bindings, context_req, protocol, options, "accept", **kwargs
    )



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import logging
import struct
import typing

from spnego._context import (
    IOV,
    ContextProxy,
    ContextReq,
    IOVUnwrapResult,
    IOVWrapResult,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
    split_username,
    wrap_system_error,
)
from spnego._credential import Credential, CredentialCache, Password, unify_credentials
from spnego._text import to_text
from spnego.channel_bindings import GssChannelBindings
from spnego.exceptions import InvalidCredentialError, NegotiateOptions, SpnegoError
from spnego.exceptions import WinError as NativeError
from spnego.iov import BufferType, IOVBuffer, IOVResBuffer

log = logging.getLogger(__name__)

HAS_SSPI = True
SSPI_IMP_ERR = None
try:
    from spnego._sspi_raw import ClientContextReq
    from spnego._sspi_raw import Credential as NativeCredential
    from spnego._sspi_raw import (
        CredentialUse,
        SecBufferDesc,
        SecBufferType,
        SecPkgAttr,
        SecPkgInfo,
        SecStatus,
    )
    from spnego._sspi_raw import SecurityContext as SSPISecContext
    from spnego._sspi_raw import (
        ServerContextReq,
        SSPIQoP,
        WinNTAuthIdentity,
        accept_security_context,
        acquire_credentials_handle,
        decrypt_message,
        encrypt_message,
        initialize_security_context,
        make_signature,
        query_context_attributes,
        verify_signature,
    )
    from spnego._sspi_raw.sspi import SecPkgAttrSizes
except ImportError as e:
    SSPI_IMP_ERR = str(e)
    HAS_SSPI = False
    log.debug("SSPI bindings not available, cannot use any SSPIProxy protocols: %s" % e)


def _available_protocols() -> typing.List[str]:
    """Return a list of protocols that SSPIProxy can offer."""
    if HAS_SSPI:
        return ["kerberos", "negotiate", "ntlm"]
    else:
        return []


def _create_iov_result(iov: "SecBufferDesc") -> typing.Tuple[IOVResBuffer, ...]:
    """Converts SSPI IOV buffer to generic IOVBuffer result."""
    buffers = []
    for i in iov:
        buffer_entry = IOVResBuffer(type=BufferType(i.buffer_type), data=i.buffer)
        buffers.append(buffer_entry)

    return tuple(buffers)


def _get_sspi_credential(
    principal: typing.Optional[str],
    protocol: str,
    usage: str,
    credentials: typing.List[Credential],
) -> "NativeCredential":
    """Get the SSPI credential.

    Will get an SSPI credential for the protocol specified. Currently only
    supports Password or CredentialCache credential types.

    Args:
        principal: The principal to use for the AcquireCredentialsHandle call
        protocol: The protocol of the credential.
        usage: Either `initiate` for a client context or `accept` for a server
            context.
        credentials: List of credentials to retrieve from.

    Returns:
        NativeCredential: The handle to the SSPI credential to use.
    """
    credential_kwargs: typing.Dict[str, typing.Any] = {
        "package": protocol,
        "principal": principal,
        "credential_use": CredentialUse.outbound if usage == "initiate" else CredentialUse.inbound,
    }

    for cred in credentials:
        if isinstance(cred, Password):
            domain, username = split_username(cred.username)
            credential_kwargs["auth_data"] = WinNTAuthIdentity(
                to_text(username, nonstring="passthru"),
                to_text(domain, nonstring="passthru"),
                to_text(cred.password, nonstring="passthru"),
            )

            return acquire_credentials_handle(**credential_kwargs)

        elif isinstance(cred, CredentialCache):
            return acquire_credentials_handle(**credential_kwargs)

    raise InvalidCredentialError(context_msg="No applicable credentials available")


class SSPIProxy(ContextProxy):
    """SSPI proxy class for pure SSPI on Windows.

    This proxy class for SSPI exposes this library into a common interface for SPNEGO authentication. This context
    uses compiled C code to interface directly into the SSPI functions on Windows to provide a native SPNEGO
    implementation.
    """

    def __init__(
        self,
        username: typing.Optional[typing.Union[str, Credential, typing.List[Credential]]] = None,
        password: typing.Optional[str] = None,
        hostname: typing.Optional[str] = None,
        service: typing.Optional[str] = None,
        channel_bindings: typing.Optional[GssChannelBindings] = None,
        context_req: ContextReq = ContextReq.default,
        usage: str = "initiate",
        protocol: str = "negotiate",
        options: NegotiateOptions = NegotiateOptions.none,
        **kwargs: typing.Any,
    ) -> None:

        if not HAS_SSPI:
            raise ImportError("SSPIProxy requires the SSPI Cython extension to be compiled: %s" % SSPI_IMP_ERR)

        credentials = unify_credentials(username, password)
        super(SSPIProxy, self).__init__(
            credentials, hostname, service, channel_bindings, context_req, usage, protocol, options
        )

        self._block_size = 0
        self._max_signature = 0
        self._security_trailer = 0

        self._complete = False
        self._context = SSPISecContext()
        self.__seq_num = 0

        sspi_credential = kwargs.get("_sspi_credential", None)
        if not sspi_credential:
            try:
                principal = self.spn if usage == "accept" else None
                sspi_credential = _get_sspi_credential(principal, protocol, usage, credentials)
            except NativeError as win_err:
                raise SpnegoError(base_error=win_err, context_msg="Getting SSPI credential") from win_err

        self._credential = sspi_credential

    @classmethod
    def available_protocols(cls, options: typing.Optional[NegotiateOptions] = None) -> typing.List[str]:
        return _available_protocols()

    @property
    def client_principal(self) -> typing.Optional[str]:
        if self.usage == "accept":
            return typing.cast(str, query_context_attributes(self._context, SecPkgAttr.names))
        else:
            return None

    @property
    def complete(self) -> bool:
        return self._complete

    @property
    def negotiated_protocol(self) -> typing.Optional[str]:
        # FIXME: Try and replicate GSSAPI. Will return None for acceptor until the first token is returned. Negotiate
        # for both iniator and acceptor until the context is established.
        package_info = typing.cast(SecPkgInfo, query_context_attributes(self._context, SecPkgAttr.package_info))
        return package_info.name.lower()

    @property
    @wrap_system_error(NativeError, "Retrieving session key")
    def session_key(self) -> bytes:
        return typing.cast(bytes, query_context_attributes(self._context, SecPkgAttr.session_key))

    def new_context(self) -> "SSPIProxy":
        return SSPIProxy(
            hostname=self._hostname,
            service=self._service,
            channel_bindings=self.channel_bindings,
            context_req=self.context_req,
            usage=self.usage,
            protocol=self.protocol,
            options=self.options,
            _sspi_credential=self._credential,
        )

    @wrap_system_error(NativeError, "Processing security token")
    def step(
        self,
        in_token: typing.Optional[bytes] = None,
        *,
        channel_bindings: typing.Optional[GssChannelBindings] = None,
    ) -> typing.Optional[bytes]:
        if not self._is_wrapped:
            log.debug("SSPI step input: %s", base64.b64encode(in_token or b"").decode())

        sec_tokens: typing.List[typing.Tuple[int, bytes]] = []
        if in_token:
            in_token = bytearray(in_token)
            sec_tokens.append((SecBufferType.token, in_token))

        if not channel_bindings:
            channel_bindings = self.channel_bindings

        if channel_bindings:
            sec_tokens.append((SecBufferType.channel_bindings, self._get_native_bindings(channel_bindings)))

        in_buffer: typing.Optional[SecBufferDesc] = None
        if sec_tokens:
            in_buffer = SecBufferDesc(len(sec_tokens))
            for idx, sec_token in enumerate(sec_tokens):
                sec_buffer = in_buffer[idx]
                sec_buffer.buffer_type = sec_token[0]
                sec_buffer.buffer = sec_token[1]

        out_buffer = SecBufferDesc(1)
        out_buffer[0].buffer_type = SecBufferType.token

        if self.usage == "initiate":
            context_req = self._context_req | ClientContextReq.allocate_memory
            res = initialize_security_context(
                self._credential,
                self._context,
                self.spn or "",
                context_req=context_req,
                input_buffer=in_buffer,
                output_buffer=out_buffer,
            )
        else:
            context_req = self._context_req | ServerContextReq.allocate_memory
            res = accept_security_context(
                self._credential,
                self._context,
                in_buffer,
                context_req=context_req,
                output_buffer=out_buffer,
            )

        out_token: typing.Optional[bytes] = None
        try:
            raw_bytes = out_buffer[0].buffer
            out_token = bytes(raw_bytes) if raw_bytes else None
        finally:
            out_buffer[0].free()

        self._context_attr = int(self._context.context_attr)

        if res == SecStatus.SEC_E_OK:
            self._complete = True

            attr_sizes = typing.cast(SecPkgAttrSizes, query_context_attributes(self._context, SecPkgAttr.sizes))
            self._block_size = attr_sizes.block_size
            self._max_signature = attr_sizes.max_signature
            self._security_trailer = attr_sizes.security_trailer

        if not self._is_wrapped:
            log.debug("SSPI step output: %s", base64.b64encode(out_token or b"").decode())

        return out_token

    def wrap(self, data: bytes, encrypt: bool = True, qop: typing.Optional[int] = None) -> WrapResult:
        res = self.wrap_iov([BufferType.header, data, BufferType.padding], encrypt=encrypt, qop=qop)
        return WrapResult(data=b"".join([r.data for r in res.buffers if r.data]), encrypted=res.encrypted)

    @wrap_system_error(NativeError, "Wrapping IOV buffer")
    def wrap_iov(
        self,
        iov: typing.Iterable[IOV],
        encrypt: bool = True,
        qop: typing.Optional[int] = None,
    ) -> IOVWrapResult:
        qop = qop or 0
        if encrypt and qop & SSPIQoP.wrap_no_encrypt:
            raise ValueError("Cannot set qop with SECQOP_WRAP_NO_ENCRYPT and encrypt=True")
        elif not encrypt:
            qop |= SSPIQoP.wrap_no_encrypt

        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = SecBufferDesc(len(buffers))
        for idx, buffer in enumerate(buffers):
            sec_buffer = iov_buffer[idx]
            sec_buffer.buffer_type = buffer[0]
            sec_buffer.buffer = buffer[1]

        encrypt_message(self._context, iov_buffer, seq_no=self._seq_num, qop=qop)

        return IOVWrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypt)

    def wrap_winrm(self, data: bytes) -> WinRMWrapResult:
        iov = self.wrap_iov([BufferType.header, data]).buffers
        header = iov[0].data or b""
        enc_data = iov[1].data or b""

        return WinRMWrapResult(header=header, data=enc_data, padding_length=0)

    def unwrap(self, data: bytes) -> UnwrapResult:
        res = self.unwrap_iov([(BufferType.stream, data), BufferType.data])

        dec_data = res.buffers[1].data or b""
        return UnwrapResult(data=dec_data, encrypted=res.encrypted, qop=res.qop)

    @wrap_system_error(NativeError, "Unwrapping IOV buffer")
    def unwrap_iov(
        self,
        iov: typing.Iterable[IOV],
    ) -> IOVUnwrapResult:
        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = SecBufferDesc(len(buffers))
        for idx, buffer in enumerate(buffers):
            sec_buffer = iov_buffer[idx]
            sec_buffer.buffer_type = buffer[0]
            sec_buffer.buffer = buffer[1]

        qop = decrypt_message(self._context, iov_buffer, seq_no=self._seq_num)
        encrypted = qop & SSPIQoP.wrap_no_encrypt == 0

        return IOVUnwrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypted, qop=qop)

    def unwrap_winrm(self, header: bytes, data: bytes) -> bytes:
        iov = self.unwrap_iov([(BufferType.header, header), data]).buffers
        return iov[1].data or b""

    @wrap_system_error(NativeError, "Signing message")
    def sign(self, data: bytes, qop: typing.Optional[int] = None) -> bytes:
        data = bytearray(data)
        signature = bytearray(self._max_signature)
        iov = SecBufferDesc(2)
        iov[0].buffer_type = SecBufferType.data
        iov[0].buffer = data
        iov[1].buffer_type = SecBufferType.token
        iov[1].buffer = signature
        make_signature(self._context, qop or 0, iov, self._seq_num)

        return iov[1].buffer or b""

    @wrap_system_error(NativeError, "Verifying message")
    def verify(self, data: bytes, mic: bytes) -> int:
        data = bytearray(data)
        mic = bytearray(mic)
        iov = SecBufferDesc(2)
        iov[0].buffer_type = SecBufferType.data
        iov[0].buffer = data
        iov[1].buffer_type = SecBufferType.token
        iov[1].buffer = mic

        return verify_signature(self._context, iov, self._seq_num)

    @property
    def _context_attr_map(self) -> typing.List[typing.Tuple[ContextReq, int]]:
        # The flags values slightly differ for a initiate and accept context.
        sspi_req = ClientContextReq if self.usage == "initiate" else ServerContextReq

        attr_map = [
            # SSPI does not differ between delegate and delegate_policy, it always respects delegate_policy.
            (ContextReq.delegate, "delegate"),
            (ContextReq.delegate_policy, "delegate"),
            (ContextReq.mutual_auth, "mutual_auth"),
            (ContextReq.replay_detect, "replay_detect"),
            (ContextReq.sequence_detect, "sequence_detect"),
            (ContextReq.confidentiality, "confidentiality"),
            (ContextReq.integrity, "integrity"),
            (ContextReq.identify, "identify"),
        ]
        if self.usage == "initiate":
            attr_map.append((ContextReq.no_integrity, "no_integrity"))

        attrs = []
        for spnego_flag, gssapi_name in attr_map:
            attrs.append((spnego_flag, getattr(sspi_req, gssapi_name)))

        return attrs

    @property
    def _seq_num(self) -> int:
        num = self.__seq_num
        self.__seq_num += 1
        return num

    def _convert_iov_buffer(self, buffer: IOVBuffer) -> typing.Tuple[int, bytearray]:
        data = bytearray()

        if isinstance(buffer.data, bytes):
            data = bytearray(buffer.data)
        elif isinstance(buffer.data, int) and not isinstance(buffer.data, bool):
            data = bytearray(buffer.data)
        else:
            auto_alloc_size = {
                BufferType.header: self._security_trailer,
                BufferType.padding: self._block_size,
                BufferType.trailer: self._security_trailer,
            }

            # If alloc wasn't explicitly set, only alloc if the type is a specific auto alloc type.
            alloc = buffer.data
            if alloc is None:
                alloc = buffer.type in auto_alloc_size

            if alloc:
                if buffer.type not in auto_alloc_size:
                    raise ValueError(
                        "Cannot auto allocate buffer of type %s.%s" % (type(buffer.type).__name__, buffer.type.name)
                    )

                data = bytearray(auto_alloc_size[buffer.type])

        return (buffer.type, data)

    def _get_native_bindings(self, channel_bindings: GssChannelBindings) -> bytearray:
        """Gets the raw byte value of the SEC_CHANNEL_BINDINGS structure."""
        b_bindings = bytearray()
        b_bindings_data = bytearray()

        def _pack_binding(name: str, b_bindings: bytearray, b_bindings_data: bytearray) -> None:
            if name == "application":
                b_data = getattr(channel_bindings, "application_data") or b""

            else:
                b_bindings += struct.pack("<I", getattr(channel_bindings, "%s_addrtype" % name))
                b_data = getattr(channel_bindings, "%s_address" % name) or b""

            b_bindings += struct.pack("<I", len(b_data))
            b_bindings += struct.pack("I", 32 + len(b_bindings_data))
            b_bindings_data += b_data

        _pack_binding("initiator", b_bindings, b_bindings_data)
        _pack_binding("acceptor", b_bindings, b_bindings_data)
        _pack_binding("application", b_bindings, b_bindings_data)

        return b_bindings + b_bindings_data



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Kerberos Messages

Various Kerberos message structures that is used by 'python -m spnego' to unpack raw ASN.1 values and print out a
pretty structure for end users. This code is not used in the actual spnego authentication processes and is just used
for debugging purposes.
"""

import base64
import collections
import datetime
import enum
import struct
import typing

from spnego._asn1 import (
    ASN1Value,
    TagClass,
    extract_asn1_tlv,
    get_sequence_value,
    unpack_asn1,
    unpack_asn1_bit_string,
    unpack_asn1_general_string,
    unpack_asn1_generalized_time,
    unpack_asn1_integer,
    unpack_asn1_octet_string,
    unpack_asn1_sequence,
    unpack_asn1_tagged_sequence,
)
from spnego._text import to_text


def _enum_labels(
    value: typing.Union[int, str, enum.Enum],
    enum_type: typing.Optional[typing.Type] = None,
) -> typing.Dict[int, str]:
    """Gets the human friendly labels of a known enum and what value they map to."""

    def get_labels(v: typing.Any) -> typing.Dict[int, str]:
        return typing.cast(typing.Dict[int, str], getattr(v, "native_labels", lambda: {})())

    return get_labels(enum_type) if enum_type else get_labels(value)


def parse_enum(
    value: typing.Union[int, str, enum.Enum],
    enum_type: typing.Optional[typing.Type] = None,
) -> str:
    """Parses an IntEnum into a human representative object of that enum."""
    enum_name = "UNKNOWN"
    if isinstance(value, enum.Enum):
        enum_name = value.name

    labels = _enum_labels(value, enum_type)
    value = value.value if isinstance(value, enum.Enum) else value

    for v, name in labels.items():
        if value == v:
            enum_name = name
            break

    return "%s (%s)" % (enum_name, value)


def parse_flags(
    value: typing.Union[int, enum.IntFlag],
    enum_type: typing.Optional[typing.Type] = None,
) -> typing.Dict[str, typing.Any]:
    """Parses an IntFlag into each flag value that is set."""
    raw_value = int(value)
    flags = []

    labels = _enum_labels(value, enum_type)
    value = int(value)

    for v, name in labels.items():
        if value & v == v:
            value &= ~v
            flags.append("%s (%d)" % (name, v))

    if value != 0:
        flags.append("UNKNOWN (%d)" % value)

    return {
        "raw": raw_value,
        "flags": flags,
    }


def parse_kerberos_token(
    token: typing.Union["KerberosV5Msg", "PAData", "PAETypeInfo2", "EncryptedData", "Ticket", "KdcReqBody"],
    secret: typing.Optional[str] = None,
    encoding: typing.Optional[str] = None,
) -> typing.Union[str, typing.Dict[str, typing.Any]]:
    """Parses a KerberosV5Msg object to a dict."""
    text_encoding = encoding if encoding else "utf-8"

    def parse_default(value: typing.Any) -> typing.Any:
        return value

    def parse_datetime(value: datetime.datetime) -> str:
        return value.isoformat()

    def parse_text(value: bytes) -> str:
        return to_text(value, encoding=text_encoding, errors="replace")

    def parse_bytes(value: bytes) -> str:
        return base64.b16encode(value).decode()

    def parse_principal_name(value: PrincipalName) -> typing.Dict[str, typing.Any]:
        return {
            "name-type": parse_enum(value.name_type),
            "name-string": [parse_text(v) for v in value.value],
        }

    def parse_host_address(value: HostAddress) -> typing.Dict[str, typing.Any]:
        return {
            "addr-type": parse_enum(value.addr_type),
            "address": parse_text(value.value),
        }

    def parse_token(value: typing.Any) -> typing.Union[str, typing.Dict[str, typing.Any]]:
        return parse_kerberos_token(value, secret, text_encoding)

    if isinstance(token, bytes):
        return parse_bytes(token)

    msg = {}
    for name, attr_name, attr_type in getattr(token, "PARSE_MAP", {}):
        attr_value = getattr(token, attr_name)

        parse_args = []
        if isinstance(attr_type, tuple):
            parse_args.append(attr_type[1])
            attr_type = attr_type[0]

        parse_func: typing.Callable = {  # type: ignore[assignment] # No idea why
            ParseType.default: parse_default,
            ParseType.enum: parse_enum,
            ParseType.flags: parse_flags,
            ParseType.datetime: parse_datetime,
            ParseType.text: parse_text,
            ParseType.bytes: parse_bytes,
            ParseType.principal_name: parse_principal_name,
            ParseType.host_address: parse_host_address,
            ParseType.token: parse_token,
        }[attr_type]

        if attr_value is None:
            parsed_value = None

        elif isinstance(attr_value, list):
            parsed_value = [parse_func(v, *parse_args) if v is not None else None for v in attr_value]

        else:
            parsed_value = parse_func(attr_value, *parse_args)

        msg[name] = parsed_value

    return msg


def unpack_hostname(value: typing.Union[bytes, ASN1Value]) -> "HostAddress":
    """Unpacks an ASN.1 value to a HostAddress."""
    s = unpack_asn1_tagged_sequence(value)

    name_type = KerberosHostAddressType(get_sequence_value(s, 0, "HostAddress", "addr-type", unpack_asn1_integer))
    name = get_sequence_value(s, 1, "HostAddress", "address", unpack_asn1_octet_string)

    return HostAddress(name_type, name)


def unpack_principal_name(value: typing.Union[bytes, ASN1Value]) -> "PrincipalName":
    """Unpacks an ASN.1 value to a PrincipalName."""
    s = unpack_asn1_tagged_sequence(value)

    name_type = KerberosPrincipalNameType(get_sequence_value(s, 0, "PrincipalName", "name-type", unpack_asn1_integer))
    name = [
        unpack_asn1_general_string(n)
        for n in get_sequence_value(s, 1, "PrincipalName", "name-string", unpack_asn1_sequence)
    ]

    return PrincipalName(name_type, name)


HostAddress = collections.namedtuple("HostAddress", ["addr_type", "value"])
"""Kerberos HostAddress and HostAddresses

A Kerberos Host Address ASN.1 definition is defined in `RFC 4120 5.2.5`_::

    HostAddress     ::= SEQUENCE  {
        addr-type       [0] Int32,
        address         [1] OCTET STRING
    }

Attributes:
    addr_type (KerberosHostAddressType): The type of address that was encoded.
    value (bytes): The address as a byte string.

.. _RFC 4120 5.2.5:
    https://www.rfc-editor.org/rfc/rfc4120#section-5.2.5
"""

PrincipalName = collections.namedtuple("PrincipalName", ["name_type", "value"])
"""Kerberos Realm and PrincipalName

A Kerberos Principal Name ASN.1 definition is defined in `RFC 4120 5.2.2`_::

    PrincipalName   ::= SEQUENCE {
        name-type       [0] Int32,
        name-string     [1] SEQUENCE OF KerberosString
    }

Attributes:
    name_type (KerberosPrincipalNameType): The type of name the was encoded.
    value (List[bytes]): Each component that forms the name.

.. _RFC 4120 5.2.2:
    https://www.rfc-editor.org/rfc/rfc4120#section-5.2.2
"""


# https://www.rfc-editor.org/rfc/rfc4120#section-5.5.1 - ap-options
class KerberosAPOptions(enum.IntFlag):
    mutual_required = 0x00000020
    use_session_key = 0x00000040
    reserved = 0x00000080

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosAPOptions", str]:
        return {
            KerberosAPOptions.mutual_required: "mutual-required",
            KerberosAPOptions.use_session_key: "use-session-key",
            KerberosAPOptions.reserved: "reserved",
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1 - KDCOptions
class KerberosKDCOptions(enum.IntFlag):
    reserved = 0x80000000
    forwardable = 0x40000000
    forwarded = 0x20000000
    proxiable = 0x10000000
    proxy = 0x08000000
    allow_postdate = 0x04000000
    postdated = 0x02000000
    unused7 = 0x01000000
    renewable = 0x00800000
    unused9 = 0x00400000
    unused10 = 0x00200000
    opt_hardware_auth = 0x00100000
    unused12 = 0x00080000
    unused13 = 0x00040000
    constrained_delegation = 0x00020000
    canonicalize = 0x00010000
    request_anonymous = 0x00008000
    unused17 = 0x00004000
    unused18 = 0x00002000
    unused19 = 0x00001000
    unused20 = 0x00000800
    unused21 = 0x00000400
    unused22 = 0x00000200
    unused23 = 0x00000100
    unused24 = 0x00000080
    unused25 = 0x00000040
    disable_transited_check = 0x00000020
    renewable_ok = 0x00000010
    enc_tkt_in_skey = 0x00000008
    unused29 = 0x00000004
    renew = 0x00000002
    validate = 0x00000001

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosKDCOptions", str]:
        return {
            KerberosKDCOptions.reserved: "reserved",
            KerberosKDCOptions.forwardable: "forwardable",
            KerberosKDCOptions.forwarded: "forwarded",
            KerberosKDCOptions.proxiable: "proxiable",
            KerberosKDCOptions.proxy: "proxy",
            KerberosKDCOptions.allow_postdate: "allow-postdate",
            KerberosKDCOptions.postdated: "postdated",
            KerberosKDCOptions.unused7: "unused7",
            KerberosKDCOptions.renewable: "renewable",
            KerberosKDCOptions.unused9: "unused9",
            KerberosKDCOptions.unused10: "unused10",
            KerberosKDCOptions.opt_hardware_auth: "opt-hardware-auth",
            KerberosKDCOptions.unused12: "unused12",
            KerberosKDCOptions.unused13: "unused13",
            KerberosKDCOptions.constrained_delegation: "constrained-delegation",
            KerberosKDCOptions.canonicalize: "canonicalize",
            KerberosKDCOptions.request_anonymous: "request-anonymous",
            KerberosKDCOptions.unused17: "unused17",
            KerberosKDCOptions.unused18: "unused18",
            KerberosKDCOptions.unused19: "unused19",
            KerberosKDCOptions.unused20: "unused20",
            KerberosKDCOptions.unused21: "unused21",
            KerberosKDCOptions.unused22: "unused22",
            KerberosKDCOptions.unused23: "unused23",
            KerberosKDCOptions.unused24: "unused24",
            KerberosKDCOptions.unused25: "unused25",
            KerberosKDCOptions.disable_transited_check: "disable-transited-check",
            KerberosKDCOptions.renewable_ok: "renewable-ok",
            KerberosKDCOptions.enc_tkt_in_skey: "enc-tkt-in-skey",
            KerberosKDCOptions.unused29: "unused29",
            KerberosKDCOptions.renew: "renew",
            KerberosKDCOptions.validate: "validate",
        }


# https://ldapwiki.com/wiki/Kerberos%20Encryption%20Types - etypes
class KerberosEncryptionType(enum.IntEnum):
    des_cbc_crc = 0x0001
    des_cbc_md4 = 0x0002
    des_cbc_md5 = 0x0003
    des_cbc_raw = 0x0004
    des3_cbc_raw = 0x0006
    des3_cbc_sha1 = 0x0010
    aes128_cts_hmac_sha1_96 = 0x0011
    aes256_cts_hmac_sha1_96 = 0x0012
    aes128_cts_hmac_sha256_128 = 0x0013
    aes256_cts_hmac_sha384_192 = 0x0014
    rc4_hmac = 0x0017
    rc4_hmac_exp = 0x0018
    camellia128_cts_cmac = 0x0019
    camellia256_cts_cmac = 0x001A

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosEncryptionType", str]:
        return {
            KerberosEncryptionType.des_cbc_crc: "DES_CBC_CRC",
            KerberosEncryptionType.des_cbc_md4: "DES_CBC_MD4",
            KerberosEncryptionType.des_cbc_md5: "DES_CBC_MD5",
            KerberosEncryptionType.des_cbc_raw: "DES_CBC_RAW",
            KerberosEncryptionType.des3_cbc_raw: "DES3_CBC_RAW",
            KerberosEncryptionType.des3_cbc_sha1: "DES3_CBC_SHA1",
            KerberosEncryptionType.aes128_cts_hmac_sha1_96: "AES128_CTS_HMAC_SHA1_96",
            KerberosEncryptionType.aes256_cts_hmac_sha1_96: "AES256_CTS_HMAC_SHA1_96",
            KerberosEncryptionType.aes128_cts_hmac_sha256_128: "AES128_CTS_HMAC_SHA256_128",
            KerberosEncryptionType.aes256_cts_hmac_sha384_192: "AES256_CTS_HMAC_SHA384_192",
            KerberosEncryptionType.rc4_hmac: "RC4_HMAC",
            KerberosEncryptionType.rc4_hmac_exp: "RC4_HMAC_EXP",
            KerberosEncryptionType.camellia128_cts_cmac: "CAMELLIA128_CTS_CMAC",
            KerberosEncryptionType.camellia256_cts_cmac: "CAMELLIA256_CTS_CMAC",
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-7.5.9
class KerberosErrorCode(enum.IntEnum):
    none = 0
    name_exp = 1
    service_exp = 2
    bad_pvno = 3
    c_old_mast_kvno = 4
    s_old_mast_kvno = 5
    c_principal_unknown = 6
    s_principal_unknown = 7
    principal_not_unique = 8
    null_key = 9
    cannot_postdate = 10
    never_valid = 11
    policy = 12
    badoption = 13
    etype_nosupp = 14
    sumtype_nosupp = 15
    padata_type_nosupp = 16
    trtype_nosupp = 17
    client_revoked = 18
    service_revoked = 19
    tgt_revoked = 20
    client_notyet = 21
    service_notyet = 22
    key_expired = 23
    preauth_failed = 24
    preauth_required = 25
    server_nomatch = 26
    must_use_user2user = 27
    path_not_accepted = 28
    kdc_svc_unavailable = 29
    ap_bad_integrity = 31
    ap_txt_expired = 32
    ap_tkt_nyv = 33
    ap_repeat = 34
    ap_not_use = 35
    ap_badmatch = 36
    ap_skew = 37
    ap_badaddr = 38
    ap_badversion = 39
    ap_msg_type = 40
    ap_modified = 41
    ap_badorder = 42
    ap_badkeyver = 44
    ap_nokey = 45
    ap_mut_fail = 46
    ap_baddirection = 47
    ap_method = 48
    ap_badseq = 49
    ap_inapp_cksum = 50
    ap_path_not_accepted = 51
    response_too_big = 52
    generic = 60
    field_toolong = 61
    kdc_client_not_trusted = 62
    kdc_not_trusted = 53
    kdc_invalid_sig = 64
    kdc_key_too_weak = 65
    kdc_certificate_mismatch = 66
    ap_no_tgt = 67
    kdc_wrong_realm = 68
    ap_user_to_user_required = 69
    kdc_cant_verify_certificate = 70
    kdc_invalid_certificate = 71
    kdc_revoked_certificate = 72
    kdc_revocation_status_unknown = 73
    kdc_revocation_status_unavailable = 74
    kdc_client_name_mismatch = 75
    kdc_name_mismatch = 76

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosErrorCode", str]:
        return {
            KerberosErrorCode.none: "KDC_ERR_NONE",
            KerberosErrorCode.name_exp: "KDC_ERR_NAME_EXP",
            KerberosErrorCode.service_exp: "KDC_ERR_SERVICE_EXP",
            KerberosErrorCode.bad_pvno: "KDC_ERR_BAD_PVNO",
            KerberosErrorCode.c_old_mast_kvno: "KDC_ERR_C_OLD_MAST_KVNO",
            KerberosErrorCode.s_old_mast_kvno: "KDC_ERR_S_OLD_MAST_KVNO",
            KerberosErrorCode.c_principal_unknown: "KDC_ERR_C_PRINCIPAL_UNKNOWN",
            KerberosErrorCode.s_principal_unknown: "KDC_ERR_S_PRINCIPAL_UNKNOWN",
            KerberosErrorCode.principal_not_unique: "KDC_ERR_PRINCIPAL_NOT_UNIQUE",
            KerberosErrorCode.null_key: "KDC_ERR_NULL_KEY",
            KerberosErrorCode.cannot_postdate: "KDC_ERR_CANNOT_POSTDATE",
            KerberosErrorCode.never_valid: "KDC_ERR_NEVER_VALID",
            KerberosErrorCode.policy: "KDC_ERR_POLICY",
            KerberosErrorCode.badoption: "KDC_ERR_BADOPTION",
            KerberosErrorCode.etype_nosupp: "KDC_ERR_ETYPE_NOSUPP",
            KerberosErrorCode.sumtype_nosupp: "KDC_ERR_SUMTYPE_NOSUPP",
            KerberosErrorCode.padata_type_nosupp: "KDC_ERR_PADATA_TYPE_NOSUPP",
            KerberosErrorCode.trtype_nosupp: "KDC_ERR_TRTYPE_NOSUPP",
            KerberosErrorCode.client_revoked: "KDC_ERR_CLIENT_REVOKED",
            KerberosErrorCode.service_revoked: "KDC_ERR_SERVICE_REVOKED",
            KerberosErrorCode.tgt_revoked: "KDC_ERR_TGT_REVOKED",
            KerberosErrorCode.client_notyet: "KDC_ERR_CLIENT_NOTYET",
            KerberosErrorCode.service_notyet: "KDC_ERR_SERVICE_NOTYET",
            KerberosErrorCode.key_expired: "KDC_ERR_KEY_EXPIRED",
            KerberosErrorCode.preauth_failed: "KDC_ERR_PREAUTH_FAILED",
            KerberosErrorCode.preauth_required: "KDC_ERR_PREAUTH_REQUIRED",
            KerberosErrorCode.server_nomatch: "KDC_ERR_SERVER_NOMATCH",
            KerberosErrorCode.must_use_user2user: "KDC_ERR_MUST_USE_USER2USER",
            KerberosErrorCode.path_not_accepted: "KDC_ERR_PATH_NOT_ACCEPTED",
            KerberosErrorCode.kdc_svc_unavailable: "KDC_ERR_SVC_UNAVAILABLE",
            KerberosErrorCode.ap_bad_integrity: "KRB_AP_ERR_BAD_INTEGRITY",
            KerberosErrorCode.ap_txt_expired: "KRB_AP_ERR_TKT_EXPIRED",
            KerberosErrorCode.ap_tkt_nyv: "KRB_AP_ERR_TKT_NYV",
            KerberosErrorCode.ap_repeat: "KRB_AP_ERR_REPEAT",
            KerberosErrorCode.ap_not_use: "KRB_AP_ERR_NOT_US",
            KerberosErrorCode.ap_badmatch: "KRB_AP_ERR_BADMATCH",
            KerberosErrorCode.ap_skew: "KRB_AP_ERR_SKEW",
            KerberosErrorCode.ap_badaddr: "KRB_AP_ERR_BADADDR",
            KerberosErrorCode.ap_badversion: "KRB_AP_ERR_BADVERSION",
            KerberosErrorCode.ap_msg_type: "KRB_AP_ERR_MSG_TYPE",
            KerberosErrorCode.ap_modified: "KRB_AP_ERR_MODIFIED",
            KerberosErrorCode.ap_badorder: "KRB_AP_ERR_BADORDER",
            KerberosErrorCode.ap_badkeyver: "KRB_AP_ERR_BADKEYVER",
            KerberosErrorCode.ap_nokey: "KRB_AP_ERR_NOKEY",
            KerberosErrorCode.ap_mut_fail: "KRB_AP_ERR_MUT_FAIL",
            KerberosErrorCode.ap_baddirection: "KRB_AP_ERR_BADDIRECTION",
            KerberosErrorCode.ap_method: "KRB_AP_ERR_METHOD",
            KerberosErrorCode.ap_badseq: "KRB_AP_ERR_BADSEQ",
            KerberosErrorCode.ap_inapp_cksum: "KRB_AP_ERR_INAPP_CKSUM",
            KerberosErrorCode.ap_path_not_accepted: "KRB_AP_PATH_NOT_ACCEPTED",
            KerberosErrorCode.response_too_big: "KRB_ERR_RESPONSE_TOO_BIG",
            KerberosErrorCode.generic: "KRB_ERR_GENERIC",
            KerberosErrorCode.field_toolong: "KRB_ERR_FIELD_TOOLONG",
            KerberosErrorCode.kdc_client_not_trusted: "KDC_ERROR_CLIENT_NOT_TRUSTED",
            KerberosErrorCode.kdc_not_trusted: "KDC_ERROR_KDC_NOT_TRUSTED",
            KerberosErrorCode.kdc_invalid_sig: "KDC_ERROR_INVALID_SIG",
            KerberosErrorCode.kdc_key_too_weak: "KDC_ERR_KEY_TOO_WEAK",
            KerberosErrorCode.kdc_certificate_mismatch: "KDC_ERR_CERTIFICATE_MISMATCH",
            KerberosErrorCode.ap_no_tgt: "KRB_AP_ERR_NO_TGT",
            KerberosErrorCode.kdc_wrong_realm: "KDC_ERR_WRONG_REALM",
            KerberosErrorCode.ap_user_to_user_required: "KRB_AP_ERR_USER_TO_USER_REQUIRED",
            KerberosErrorCode.kdc_cant_verify_certificate: "KDC_ERR_CANT_VERIFY_CERTIFICATE",
            KerberosErrorCode.kdc_invalid_certificate: "KDC_ERR_INVALID_CERTIFICATE",
            KerberosErrorCode.kdc_revoked_certificate: "KDC_ERR_REVOKED_CERTIFICATE",
            KerberosErrorCode.kdc_revocation_status_unknown: "KDC_ERR_REVOCATION_STATUS_UNKNOWN",
            KerberosErrorCode.kdc_revocation_status_unavailable: "KDC_ERR_REVOCATION_STATUS_UNAVAILABLE",
            KerberosErrorCode.kdc_client_name_mismatch: "KDC_ERR_CLIENT_NAME_MISMATCH",
            KerberosErrorCode.kdc_name_mismatch: "KDC_ERR_KDC_NAME_MISMATCH",
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-5.10
class KerberosMessageType(enum.IntEnum):
    unknown = 0
    as_req = 10
    as_rep = 11
    tgs_req = 12
    tgs_rep = 13
    ap_req = 14
    ap_rep = 15
    error = 30

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosMessageType", str]:
        return {
            KerberosMessageType.unknown: "UNKNOWN",
            KerberosMessageType.as_req: "AS-REQ",
            KerberosMessageType.as_rep: "AS-REP",
            KerberosMessageType.tgs_req: "TGS-REQ",
            KerberosMessageType.tgs_rep: "TGS-REP",
            KerberosMessageType.ap_req: "AP-REQ",
            KerberosMessageType.ap_rep: "AP-REP",
            KerberosMessageType.error: "KRB-ERROR",
        }


# https://www.iana.org/assignments/kerberos-parameters/kerberos-parameters.xhtml
class KerberosPADataType(enum.IntEnum):
    tgs_req = 1
    enc_timestamp = 2
    pw_salt = 3
    reserved = 4
    enc_unix_time = 5
    sandia_secureid = 6
    sesame = 7
    osf_dce = 8
    cybersafe_secureid = 9
    afs3_salt = 10
    etype_info = 11
    sam_challenge = 12
    sam_response = 13
    pk_as_req_old = 14
    pk_as_rep_old = 15
    pk_as_req = 16
    pk_as_rep = 17
    pk_ocsp_response = 18
    etype_info2 = 19
    use_specified_kvno = 20
    svr_referral_info = 20
    sam_redirect = 21
    get_from_typed_data = 22
    td_padata = 22
    sam_etype_info = 23
    alt_princ = 24
    server_referral = 25
    sam_challenge2 = 30
    sam_response2 = 31
    extra_tgt = 41
    td_pkinit_cms_certificates = 101
    td_krb_principal = 102
    td_krb_realm = 103
    td_trusted_certifiers = 104
    td_certificate_index = 105
    td_app_defined_error = 106
    td_req_nonce = 107
    td_req_seq = 108
    td_dh_parameters = 109
    td_cms_digest_algorithms = 111
    td_cert_digest_algorithms = 112
    pac_request = 128
    for_user = 128
    for_x509_user = 130
    for_check_dups = 131
    as_checksum = 132
    fx_cookie = 133
    authentication_set = 134
    auth_set_selected = 165
    fx_fast = 136
    fx_error = 137
    encrypted_challenge = 138
    otp_challenge = 141
    otp_request = 142
    otp_confirm = 143
    otp_pin_change = 144
    epak_as_req = 145
    epak_as_rep = 146
    pkinit_kx = 147
    pku2u_name = 148
    enc_pa_rep = 149
    as_freshness = 150
    spake = 151
    kerb_key_list_req = 161
    kerb_key_list_rep = 162
    supported_etypes = 165
    extended_error = 166
    pac_options = 167

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosPADataType", str]:
        return {
            KerberosPADataType.tgs_req: "PA-TGS-REQ",
            KerberosPADataType.enc_timestamp: "PA-ENC-TIMESTAMP",
            KerberosPADataType.pw_salt: "PA-PW-SALT",
            KerberosPADataType.reserved: "reserved",
            KerberosPADataType.enc_unix_time: "PA-ENC-UNIX-TIME",
            KerberosPADataType.sandia_secureid: "PA-SANDIA-SECUREID",
            KerberosPADataType.sesame: "PA-SESAME",
            KerberosPADataType.osf_dce: "PA-OSF-DCE",
            KerberosPADataType.cybersafe_secureid: "PA-CYBERSAFE-SECUREID",
            KerberosPADataType.afs3_salt: "PA-AFS3-SALT",
            KerberosPADataType.etype_info: "PA-ETYPE-INFO",
            KerberosPADataType.sam_challenge: "PA-SAM-CHALLENGE",
            KerberosPADataType.sam_response: "PA-SAM-RESPONSE",
            KerberosPADataType.pk_as_req_old: "PA-PK-AS-REQ_OLD",
            KerberosPADataType.pk_as_rep_old: "PA-PK-AS-REP_OLD",
            KerberosPADataType.pk_as_req: "PA-PK-AS-REQ",
            KerberosPADataType.pk_as_rep: "PA-PK-AS-REP",
            KerberosPADataType.pk_ocsp_response: "PA-PK-OCSP-RESPONSE",
            KerberosPADataType.etype_info2: "PA-ETYPE-INFO2",
            KerberosPADataType.use_specified_kvno: "PA-USE-SPECIFIED-KVNO or PA-SVR-REFERRAL-INFO",
            KerberosPADataType.sam_redirect: "PA-SAM-REDIRECT",
            KerberosPADataType.get_from_typed_data: "PA-GET-FROM-TYPED-DATA",
            KerberosPADataType.td_padata: "TD-PADATA",
            KerberosPADataType.sam_etype_info: "PA-SAM-ETYPE-INFO",
            KerberosPADataType.alt_princ: "PA-ALT-PRINC",
            KerberosPADataType.server_referral: "PA-SERVER-REFERRAL",
            KerberosPADataType.sam_challenge2: "PA-SAM-CHALLENGE2",
            KerberosPADataType.sam_response2: "PA-SAM-RESPONSE2",
            KerberosPADataType.extra_tgt: "PA-EXTRA-TGT",
            KerberosPADataType.td_pkinit_cms_certificates: "TD-PKINIT-CMS-CERTIFICATES",
            KerberosPADataType.td_krb_principal: "TD-KRB-PRINCIPAL",
            KerberosPADataType.td_krb_realm: "TD-KRB-REALM",
            KerberosPADataType.td_trusted_certifiers: "TD-TRUSTED-CERTIFIERS",
            KerberosPADataType.td_certificate_index: "TD-CERTIFICATE-INDEX",
            KerberosPADataType.td_app_defined_error: "TD-APP-DEFINED-ERROR",
            KerberosPADataType.td_req_nonce: "TD-REQ-NONCE",
            KerberosPADataType.td_req_seq: "TD-REQ-SEQ",
            KerberosPADataType.td_dh_parameters: "TD_DH_PARAMETERS",
            KerberosPADataType.td_cms_digest_algorithms: "TD-CMS-DIGEST-ALGORITHMS",
            KerberosPADataType.td_cert_digest_algorithms: "TD-CERT-DIGEST-ALGORITHMS",
            KerberosPADataType.pac_request: "PA-PAC-REQUEST",
            KerberosPADataType.for_user: "PA-FOR_USER",
            KerberosPADataType.for_x509_user: "PA-FOR-X509-USER",
            KerberosPADataType.for_check_dups: "PA-FOR-CHECK_DUPS",
            KerberosPADataType.as_checksum: "PA-AS-CHECKSUM",
            KerberosPADataType.fx_cookie: "PA-FX-COOKIE",
            KerberosPADataType.authentication_set: "PA-AUTHENTICATION-SET",
            KerberosPADataType.auth_set_selected: "PA-AUTH-SET-SELECTED",
            KerberosPADataType.fx_fast: "PA-FX-FAST",
            KerberosPADataType.fx_error: "PA-FX-ERROR",
            KerberosPADataType.encrypted_challenge: "PA-ENCRYPTED-CHALLENGE",
            KerberosPADataType.otp_challenge: "PA-OTP-CHALLENGE",
            KerberosPADataType.otp_request: "PA-OTP-REQUEST",
            KerberosPADataType.otp_confirm: "PA-OTP-CONFIRM",
            KerberosPADataType.otp_pin_change: "PA-OTP-PIN-CHANGE",
            KerberosPADataType.epak_as_req: "PA-EPAK-AS-REQ",
            KerberosPADataType.epak_as_rep: "PA-EPAK-AS-REP",
            KerberosPADataType.pkinit_kx: "PA_PKINIT_KX",
            KerberosPADataType.pku2u_name: "PA_PKU2U_NAME",
            KerberosPADataType.enc_pa_rep: "PA-REQ-ENC-PA-REP",
            KerberosPADataType.as_freshness: "PA_AS_FRESHNESS",
            KerberosPADataType.spake: "PA-SPAKE",
            KerberosPADataType.kerb_key_list_req: "KERB-KEY-LIST-REQ",
            KerberosPADataType.kerb_key_list_rep: "KERB-KEY-LIST-REP",
            KerberosPADataType.supported_etypes: "PA-SUPPORTED-ETYPES",
            KerberosPADataType.extended_error: "PA-EXTENDED_ERROR",
            KerberosPADataType.pac_options: "PA-PAC-OPTIONS",
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-6.2
class KerberosPrincipalNameType(enum.IntEnum):
    unknown = 0
    principal = 1
    srv_inst = 2
    srv_hst = 3
    srv_xhst = 4
    uid = 5
    x500_principal = 6
    smtp_name = 7
    enterprise = 10

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosPrincipalNameType", str]:
        return {
            KerberosPrincipalNameType.unknown: "NT-UNKNOWN",
            KerberosPrincipalNameType.principal: "NT-PRINCIPAL",
            KerberosPrincipalNameType.srv_inst: "NT-SRV-INST",
            KerberosPrincipalNameType.srv_hst: "NT-SRV-HST",
            KerberosPrincipalNameType.srv_xhst: "NT-SRV-XHST",
            KerberosPrincipalNameType.uid: "NT-UID",
            KerberosPrincipalNameType.x500_principal: "NT-X500-PRINCIPAL",
            KerberosPrincipalNameType.smtp_name: "NT-SMTP-NAME",
            KerberosPrincipalNameType.enterprise: "NT-ENTERPRISE",
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-7.5.3
class KerberosHostAddressType(enum.IntEnum):
    ipv4 = 2
    directional = 3
    chaos_net = 5
    xns = 6
    iso = 7
    decnet_phase_iv = 12
    apple_talk_ddp = 16
    netbios = 20
    ipv6 = 26

    @classmethod
    def native_labels(cls) -> typing.Dict["KerberosHostAddressType", str]:
        return {
            KerberosHostAddressType.ipv4: "IPv4",
            KerberosHostAddressType.directional: "Directional",
            KerberosHostAddressType.chaos_net: "ChaosNet",
            KerberosHostAddressType.xns: "XNS",
            KerberosHostAddressType.iso: "ISO",
            KerberosHostAddressType.decnet_phase_iv: "DECNET Phase IV",
            KerberosHostAddressType.apple_talk_ddp: "AppleTalk DDP",
            KerberosHostAddressType.netbios: "NetBios",
            KerberosHostAddressType.ipv6: "IPv6",
        }


class ParseType(enum.IntEnum):
    default = 0
    enum = 1
    flags = 2
    datetime = 3
    text = 4
    bytes = 5
    principal_name = 6
    host_address = 7
    token = 8


class _KerberosMsgType(type):
    __registry: typing.Dict[int, typing.Dict[int, "_KerberosMsgType"]] = {}

    def __init__(
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        pvno = getattr(cls, "PVNO", 0)

        if pvno not in cls.__registry:
            cls.__registry[pvno] = {}

        msg_type = getattr(cls, "MESSAGE_TYPE", None)
        if msg_type is not None:
            cls.__registry[pvno][msg_type] = cls

    def __call__(
        cls,
        sequence: typing.Dict[int, ASN1Value],
    ) -> "_KerberosMsgType":
        # The KrbAsReq msg starts at 1 not 0, so do a check for that.
        pvno_idx = 0
        if 0 not in sequence:
            pvno_idx = 1

        pvno = unpack_asn1_integer(sequence[pvno_idx])
        message_type = unpack_asn1_integer(sequence[pvno_idx + 1])
        new_cls = cls.__registry[pvno].get(message_type, cls)
        return super(_KerberosMsgType, new_cls).__call__(sequence)


class KerberosV5Msg(metaclass=_KerberosMsgType):

    MESSAGE_TYPE = KerberosMessageType.unknown
    PVNO = 5

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        # This is only used if decoding an unknown Kerberos message type.
        self.sequence = sequence

    @staticmethod
    def unpack(value: typing.Union[ASN1Value, bytes]) -> "KerberosV5Msg":
        msg_sequence = unpack_asn1_tagged_sequence(value)
        return KerberosV5Msg(msg_sequence)


class KrbAsReq(KerberosV5Msg):
    """The KRB_AS_REQ message.

    The KRB_AS_REQ message is used when the client wishes to retrieve a the initial ticket for a service. The
    KRB_TGS_REQ message is identical except for the tag and msg-type is used when retrieving additional tickets for a
    service.

    The ASN.1 definition for the KDC-REQ structure is defined in `RFC 4120 5.4.1`_::

        KDC-REQ         ::= SEQUENCE {
            -- NOTE: first tag is [1], not [0]
            pvno            [1] INTEGER (5) ,
            msg-type        [2] INTEGER (10 -- AS -- | 12 -- TGS --),
            padata          [3] SEQUENCE OF PA-DATA OPTIONAL
                                -- NOTE: not empty --,
            req-body        [4] KDC-REQ-BODY
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        padata (PAData): The pre-authentication data.
        req_body (KdcReqBody): The body of the request.

    .. _RFC 4120 5.4.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1
    """

    MESSAGE_TYPE = KerberosMessageType.as_req

    PARSE_MAP = [
        ("pvno", "PVNO", ParseType.default),
        ("msg-type", "MESSAGE_TYPE", ParseType.enum),
        ("padata", "padata", ParseType.token),
        ("req-body", "req_body", ParseType.token),
    ]

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        def unpack_padata(value: typing.Union[ASN1Value, bytes]) -> typing.List:
            return [PAData.unpack(p) for p in unpack_asn1_sequence(value)]

        self.padata = get_sequence_value(sequence, 3, "KDC-REQ", "pa-data", unpack_padata)
        self.req_body = get_sequence_value(sequence, 4, "KDC-REQ", "req-body", KdcReqBody.unpack)


class KrbAsRep(KerberosV5Msg):
    """The KRB_AS_REP message.

    The KRB_AS_REP message is used for a reply from the KDC to a KRB_AS_REQ message. The KRB_TGS_REP message is
    identical except for the tag and msg-type.

    The ASN.1 definition for the KDC-REP structure is defined in `RFC 4120 5.4.2`_::

        KDC-REP         ::= SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (11 -- AS -- | 13 -- TGS --),
            padata          [2] SEQUENCE OF PA-DATA OPTIONAL
                                -- NOTE: not empty --,
            crealm          [3] Realm,
            cname           [4] PrincipalName,
            ticket          [5] Ticket,
            enc-part        [6] EncryptedData
                                -- EncASRepPart or EncTGSRepPart,
                                -- as appropriate
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        padata (PAData): The pre-authentication data.
        crealm (bytes): The client realm.
        cname (PrincipalName): The client principal name.
        ticket (Ticket): The newly issued ticket.
        enc_part (EncryptedData): The encrypted part of the message.

    .. _RFC 4120 5.4.2:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.2
    """

    MESSAGE_TYPE = KerberosMessageType.as_rep

    PARSE_MAP = [
        ("pvno", "PVNO", ParseType.default),
        ("msg-type", "MESSAGE_TYPE", ParseType.enum),
        ("padata", "padata", ParseType.token),
        ("crealm", "crealm", ParseType.text),
        ("cname", "cname", ParseType.principal_name),
        ("ticket", "ticket", ParseType.token),
        ("enc-part", "enc_part", ParseType.token),
    ]

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        def unpack_padata(value: typing.Union[ASN1Value, bytes]) -> typing.List:
            return [PAData.unpack(p) for p in unpack_asn1_sequence(value)]

        self.padata = get_sequence_value(sequence, 2, "KDC-REP", "pa-data", unpack_padata)
        self.crealm = get_sequence_value(sequence, 3, "KDC-REP", "crealm", unpack_asn1_general_string)
        self.cname = get_sequence_value(sequence, 4, "KDC-REP", "cname", unpack_principal_name)
        self.ticket = get_sequence_value(sequence, 5, "KDC-REP", "ticket", Ticket.unpack)
        self.enc_part = get_sequence_value(sequence, 6, "KDC-REP", "enc-part", EncryptedData.unpack)


class KrbTgsReq(KrbAsReq):
    """The KRB_TGS_REQ is the same as KRB_AS_REQ but with a different MESSAGE_TYPE."""

    MESSAGE_TYPE = KerberosMessageType.tgs_req


class KrbTgsRep(KrbAsRep):
    """The KRB_TGS_REP is the same as KRB_AS_REP but with a different MESSAGE_TYPE."""

    MESSAGE_TYPE = KerberosMessageType.tgs_rep


class KrbApReq(KerberosV5Msg):
    """The KRB_AP_REQ message.

    The KRB_AP_REQ message contains is used to authenticate the initiator to an acceptor.

    The ASN.1 definition for the KRB_AP_REQ structure is defined in `RFC 4120 5.5.1`_::

        AP-REQ          ::= [APPLICATION 14] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (14),
            ap-options      [2] APOptions,
            ticket          [3] Ticket,
            authenticator   [4] EncryptedData -- Authenticator
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        ap_options (KerberosAPOptions): Options related to the AP request.
        ticket (Ticket): The ticket authenticating the client to the server.
        authenticator (EncryptedData): The encrypted authenticator.

    .. _RFC 4120 5.5.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.5.1
    """

    MESSAGE_TYPE = KerberosMessageType.ap_req

    PARSE_MAP = [
        ("pvno", "PVNO", ParseType.default),
        ("msg-type", "MESSAGE_TYPE", ParseType.enum),
        ("ap-options", "ap_options", ParseType.flags),
        ("ticket", "ticket", ParseType.token),
        ("authenticator", "authenticator", ParseType.token),
    ]

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        raw_ap_options = get_sequence_value(sequence, 2, "AP-REQ", "ap-options", unpack_asn1_bit_string)
        ap_options = KerberosAPOptions(struct.unpack("<I", raw_ap_options)[0])

        self.ap_options = ap_options
        self.ticket = get_sequence_value(sequence, 3, "AP-REQ", "ticket", Ticket.unpack)
        self.authenticator = get_sequence_value(sequence, 4, "AP-REQ", "authenticator", EncryptedData.unpack)


class KrbApRep(KerberosV5Msg):
    """The KRB_AP_REP message.

    The KRB_AP_REP is a response to an application request `KRB_AP_REQ`.

    The ASN.1 definition for the KRB_AP_REP structure is defined in `RFC 4120 5.5.2`_::

        AP-REP          ::= [APPLICATION 15] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (15),
            enc-part        [2] EncryptedData -- EncAPRepPart
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        enc_part (EncryptedData): The encrypted authenticator.

    .. _RFC 4120 5.5.2:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.5.2
    """

    MESSAGE_TYPE = KerberosMessageType.ap_rep

    PARSE_MAP = [
        ("pvno", "PVNO", ParseType.default),
        ("msg-type", "MESSAGE_TYPE", ParseType.enum),
        ("enc-part", "enc_part", ParseType.token),
    ]

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        self.enc_part = get_sequence_value(sequence, 2, "AP-REP", "enc-part", EncryptedData.unpack)


class KrbError(KerberosV5Msg):
    """The KRB_ERROR message.

    The KRB_ERROR is a message sent in the occurrence of an error.

    The ASN.1 definition for the KRB_ERROR structure is defined in `RFC 4120 5.9.1`_::

        KRB-ERROR       ::= [APPLICATION 30] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (30),
            ctime           [2] KerberosTime OPTIONAL,
            cusec           [3] Microseconds OPTIONAL,
            stime           [4] KerberosTime,
            susec           [5] Microseconds,
            error-code      [6] Int32,
            crealm          [7] Realm OPTIONAL,
            cname           [8] PrincipalName OPTIONAL,
            realm           [9] Realm -- service realm --,
            sname           [10] PrincipalName -- service name --,
            e-text          [11] KerberosString OPTIONAL,
            e-data          [12] OCTET STRING OPTIONAL
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        ctime (datetime.datetime): The current time on the client's host.
        cusec (int): The microsecond part of the client's timestamp.
        stime (datetime.datetime): The current time of the server.
        susec (int): The microsecond part of the server's timestamp.
        error_code (KerberosErrorCode): THe error code returned by the kerberos when a request fails.
        crealm (bytes): The realm that issues a ticket.
        cname (PrincipalName): The principal name in the ticket.
        realm (bytes): The service realm.
        sname (PrincipalName): The service name.
        e_text (bytes): Additional text to explain the error code.
        e_data (bytes): Additional data about the error.

    .. _RFC 4120 5.9.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.9.1
    """

    MESSAGE_TYPE = KerberosMessageType.error

    PARSE_MAP = [
        ("pvno", "PVNO", ParseType.default),
        ("msg-type", "MESSAGE_TYPE", ParseType.enum),
        ("ctime", "ctime", ParseType.datetime),
        ("cusec", "cusec", ParseType.default),
        ("stime", "stime", ParseType.datetime),
        ("susec", "susec", ParseType.default),
        ("error-code", "error_code", ParseType.enum),
        ("crealm", "crealm", ParseType.text),
        ("cname", "cname", ParseType.principal_name),
        ("realm", "realm", ParseType.text),
        ("sname", "sname", ParseType.principal_name),
        ("e-text", "e_text", ParseType.text),
        ("e-data", "e_data", ParseType.bytes),
    ]

    def __init__(self, sequence: typing.Dict[int, ASN1Value]) -> None:
        self.ctime = get_sequence_value(sequence, 2, "KRB-ERROR", "ctime", unpack_asn1_generalized_time)
        self.cusec = get_sequence_value(sequence, 3, "KRB-ERROR", "cusec", unpack_asn1_integer)
        self.stime = get_sequence_value(sequence, 4, "KRB-ERROR", "stime", unpack_asn1_generalized_time)
        self.susec = get_sequence_value(sequence, 5, "KRB-ERROR", "susec", unpack_asn1_integer)
        self.error_code = KerberosErrorCode(
            get_sequence_value(sequence, 6, "KRB-ERROR", "error-code", unpack_asn1_integer)
        )
        self.crealm = get_sequence_value(sequence, 7, "KRB-ERROR", "crealm", unpack_asn1_general_string)
        self.cname = get_sequence_value(sequence, 8, "KRB-ERROR", "cname", unpack_principal_name)
        self.realm = get_sequence_value(sequence, 9, "KRB-ERROR", "realm", unpack_asn1_general_string)
        self.sname = get_sequence_value(sequence, 10, "KRB-ERROR", "realm", unpack_principal_name)
        self.e_text = get_sequence_value(sequence, 11, "KRB-ERROR", "e-text", unpack_asn1_general_string)
        self.e_data = get_sequence_value(sequence, 12, "KRB-ERROR", "e-data", unpack_asn1_octet_string)


class PAData:
    """Kerberos PA-DATA.

    The ASN.1 definition for the PA-DATA structure is defined in `RFC 4120 5.2.7`_::

        PA-DATA         ::= SEQUENCE {
            -- NOTE: first tag is [1], not [0]
            padata-type     [1] Int32,
            padata-value    [2] OCTET STRING -- might be encoded AP-REQ
        }

    Args:
        data_type: Indicates the type of data the value represents.
        value: The PAData value, usually the DER encoding of another message.

    Attributes:
        data_type (Union[int, KerberosPADataType]): See args.
        b_value (bytes): The raw bytes of padata-value, use `value` to get a structured object of these bytes if
            available.

    .. RFC 4120 5.2.7:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.7
    """

    PARSE_MAP = [
        ("padata-type", "data_type", ParseType.enum),
        ("padata-value", "value", ParseType.token),
    ]

    def __init__(self, data_type: typing.Union[int, KerberosPADataType], value: bytes) -> None:
        self.data_type = data_type
        self.b_value = value

    @property
    def value(self) -> typing.Any:
        if self.data_type == KerberosPADataType.tgs_req:
            # Special edge case for this PA type due to how the messages require the data in unpack
            return KrbTgsReq.unpack(unpack_asn1(unpack_asn1(self.b_value)[0].b_data)[0])

        data_type_map = {
            int(KerberosPADataType.enc_timestamp): (EncryptedData.unpack, False),
            int(KerberosPADataType.etype_info2): (PAETypeInfo2.unpack, True),
        }

        if self.data_type in data_type_map:
            unpack_func, is_sequence = data_type_map[int(self.data_type)]
            b_value = unpack_asn1(self.b_value)[0]
            if is_sequence:
                # Is a SEQUENCE OF (list of entries).
                return [unpack_func(v) for v in unpack_asn1_sequence(b_value)]

            else:
                return unpack_func(b_value.b_data)

        else:
            return self.b_value

    @staticmethod
    def unpack(value: typing.Union[ASN1Value, bytes]) -> "PAData":
        sequence = unpack_asn1_tagged_sequence(value)

        def unpack_data_type(value: typing.Union[ASN1Value, bytes]) -> typing.Union[KerberosPADataType, int]:
            int_val = unpack_asn1_integer(value)

            try:
                return KerberosPADataType(int_val)
            except ValueError:
                return int_val

        data_type = get_sequence_value(sequence, 1, "PA-DATA", "padata-type", unpack_data_type)
        pa_value = get_sequence_value(sequence, 2, "PA-DATA", "padata-value", unpack_asn1_octet_string)

        return PAData(data_type, pa_value)


class PAETypeInfo2:
    """Kerberos PA-ETYPE-INFO2 container.

    The ASN.1 definition for the PA-ETYPE-INFO2 structure is defined in `RFC 4120 5.2.7.5`_::

        ETYPE-INFO2-ENTRY       ::= SEQUENCE {
            etype           [0] Int32,
            salt            [1] KerberosString OPTIONAL,
            s2kparams       [2] OCTET STRING OPTIONAL
        }

    Args:
        etype: The etype that defines the cipher used.
        salt: The used in the cipher associated with the cryptosystem.
        s2kparams: Extra params to be interpreted by the cipher associated with the cryptosystem.

    Attributes:
        etype (KerberosEncryptionType): See args.
        salt (Optional[bytes]): See args.
        s2kparams (Optional[bytes]): See args.

    .. RFC 4120 5.2.7.5:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.7.5
    """

    PARSE_MAP = [
        ("etype", "etype", ParseType.enum),
        ("salt", "salt", ParseType.bytes),
        ("s2kparams", "s2kparams", ParseType.bytes),
    ]

    def __init__(
        self,
        etype: KerberosEncryptionType,
        salt: typing.Optional[bytes],
        s2kparams: typing.Optional[bytes],
    ) -> None:
        self.etype = etype
        self.salt = salt
        self.s2kparams = s2kparams

    @staticmethod
    def unpack(value: typing.Union[ASN1Value, bytes]) -> "PAETypeInfo2":
        sequence = unpack_asn1_tagged_sequence(value)

        etype = KerberosEncryptionType(get_sequence_value(sequence, 0, "PA-ETYPE-INFO2", "etype", unpack_asn1_integer))
        salt = get_sequence_value(sequence, 1, "ETYPE-INFO2-ENTRY", "salt", unpack_asn1_general_string)
        s2kparams = get_sequence_value(sequence, 2, "ETYPE-INFO2-ENTRY", "s2kparams", unpack_asn1_octet_string)

        return PAETypeInfo2(etype, salt, s2kparams)


class EncryptedData:
    """Kerberos EncryptedData container.

    The ASN.1 definition for the EncryptedData structure is defined in `RFC 4120 5.2.9`_::

        EncryptedData   ::= SEQUENCE {
            etype   [0] Int32 -- EncryptionType --,
            kvno    [1] UInt32 OPTIONAL,
            cipher  [2] OCTET STRING -- ciphertext
        }

    Args:
        etype: The encryption algorithm that was used to encipher the cipher.
        kvno: The version number of the key under which data is encrypted. It is only present in messages encrypted
            under long lasting keys.
        cipher: The enciphered text.

    Attributes:
        etype (KerberosEncryptionType): See args.
        kvno (Optional[int]): See args.
        cipher (bytes): See args.

    .. RFC 4120 5.2.9:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.9
    """

    PARSE_MAP = [
        ("etype", "etype", ParseType.enum),
        ("kvno", "kvno", ParseType.default),
        ("cipher", "cipher", ParseType.bytes),
    ]

    def __init__(self, etype: KerberosEncryptionType, kvno: typing.Optional[int], cipher: bytes) -> None:
        self.etype = etype
        self.kvno = kvno
        self.cipher = cipher

    @staticmethod
    def unpack(value: typing.Union[bytes, ASN1Value]) -> "EncryptedData":
        sequence = unpack_asn1_tagged_sequence(value)

        etype = KerberosEncryptionType(get_sequence_value(sequence, 0, "EncryptedData", "etype", unpack_asn1_integer))
        kvno = get_sequence_value(sequence, 1, "EncryptedData", "kvno", unpack_asn1_integer)
        cipher = get_sequence_value(sequence, 2, "EncryptedData", "cipher", unpack_asn1_octet_string)

        return EncryptedData(etype, kvno, cipher)


class Ticket:
    """Kerberos Ticket.

    The ASN.1 definition for the Ticket structure is defined in `RFC 4120 5.3`_::

        Ticket          ::= [APPLICATION 1] SEQUENCE {
            tkt-vno         [0] INTEGER (5),
            realm           [1] Realm,
            sname           [2] PrincipalName,
            enc-part        [3] EncryptedData -- EncTicketPart
        }

    Args:
        tkt_vno: The version number for the ticket format.
        realm: The realm that issued a ticket.
        sname: All the name components of the server's identity.
        enc_part: The encrypted part of the ticket, it is encrypted in the key shared by Kerberos and the end server.

    Attributes:
        tkt_vno (int): See args.
        realm (bytes): See args.
        sname (List[PrincipalName]): See args.
        enc_part (EncryptedData): See args.

    .. _RFC 4120 5.3:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.3
    """

    PARSE_MAP = [
        ("tkt-vno", "tkt_vno", ParseType.default),
        ("realm", "realm", ParseType.text),
        ("sname", "sname", ParseType.principal_name),
        ("enc-part", "enc_part", ParseType.token),
    ]

    def __init__(
        self,
        tkt_vno: int,
        realm: bytes,
        sname: typing.List[PrincipalName],
        enc_part: EncryptedData,
    ) -> None:
        self.tkt_vno = tkt_vno
        self.realm = realm
        self.sname = sname
        self.enc_part = enc_part

    @staticmethod
    def unpack(value: typing.Union[ASN1Value, bytes]) -> "Ticket":
        b_data = extract_asn1_tlv(value, TagClass.application, 1)
        sequence = unpack_asn1_tagged_sequence(unpack_asn1(b_data)[0])

        tkt_vno = get_sequence_value(sequence, 0, "Ticket", "tkt-vno", unpack_asn1_integer)
        realm = get_sequence_value(sequence, 1, "Ticket", "realm", unpack_asn1_general_string)
        sname = get_sequence_value(sequence, 2, "Ticket", "sname", unpack_principal_name)
        enc_part = get_sequence_value(sequence, 3, "Ticket", "enc-part", EncryptedData.unpack)

        return Ticket(tkt_vno, realm, sname, enc_part)


class KdcReqBody:
    """The KRB_AS_REQ message.

    The KRB_AS_REQ message is used when the client wishes to retrieve a the initial ticket for a service. The
    KRB_TGS_REQ message is identical except for the tag and msg-type is used when retrieving additional tickets for a
    service.

    The ASN.1 definition for the KDC-REQ structure is defined in `RFC 4120 5.4.1`_::

        KDC-REQ-BODY    ::= SEQUENCE {
            kdc-options             [0] KDCOptions,
            cname                   [1] PrincipalName OPTIONAL
                                        -- Used only in AS-REQ --,
            realm                   [2] Realm
                                        -- Server's realm
                                        -- Also client's in AS-REQ --,
            sname                   [3] PrincipalName OPTIONAL,
            from                    [4] KerberosTime OPTIONAL,
            till                    [5] KerberosTime,
            rtime                   [6] KerberosTime OPTIONAL,
            nonce                   [7] UInt32,
            etype                   [8] SEQUENCE OF Int32 -- EncryptionType
                                        -- in preference order --,
            addresses               [9] HostAddresses OPTIONAL,
            enc-authorization-data  [10] EncryptedData OPTIONAL
                                        -- AuthorizationData --,
            additional-tickets      [11] SEQUENCE OF Ticket OPTIONAL
                                        -- NOTE: not empty
        }

    Args:
        kdc_options: Flags desired by the client and other behaviour desired.
        cname: The client name.
        realm: The realm part of the server's principal.
        sname: The service name.
        postdated_from: When the requested ticket is to be posted from.
        postdated_till: The expiration date requested by the client.
        rtime: The requested renew-till time.
        nonce: Random number generated by the client.
        etype: The desired encryption algorithm to be used in priority order.
        addresses: Addresses from which the requested ticket is to be valid.
        enc_authorization_data: Encrypted authorization data.
        additional_tickets: Additional tickets to be optionally included in a request.

    Attributes:
        kdc_options (int): See args.
        cname (Optional[PrincipalName]): See args.
        realm (bytes): See args.
        sname (PrincipalName): See args.
        postdated_from (Optional[datetime.datetime]): See args.
        postdated_till (datetime.datetime): See args.
        rtime (Optional[datetime.datetime]): See args.
        nonce (int): See args.
        etype (List[KerberosEncryptionType]): See args.
        addresses (Optional[List[HostAddress]]): See args.
        enc_authorization_data (Optional[EncryptedData]): See args.
        additional_tickets (Optional[Ticket]): See args.

    .. _RFC 4120 5.4.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1
    """

    PARSE_MAP = [
        ("kdc-options", "kdc_options", (ParseType.flags, KerberosKDCOptions)),
        ("cname", "cname", ParseType.principal_name),
        ("realm", "realm", ParseType.text),
        ("sname", "sname", ParseType.principal_name),
        ("from", "postdated_from", ParseType.datetime),
        ("till", "postdated_till", ParseType.datetime),
        ("rtime", "rtime", ParseType.datetime),
        ("nonce", "nonce", ParseType.default),
        ("etype", "etype", ParseType.enum),
        ("addresses", "addresses", ParseType.host_address),
        ("enc-authorization-data", "enc_authorization_data", ParseType.token),
        ("additional-tickets", "additional_tickets", ParseType.token),
    ]

    def __init__(
        self,
        kdc_options: int,
        cname: typing.Optional[PrincipalName],
        realm: bytes,
        sname: PrincipalName,
        postdated_from: typing.Optional[datetime.datetime],
        postdated_till: datetime.datetime,
        rtime: typing.Optional[datetime.datetime],
        nonce: int,
        etype: typing.List[KerberosEncryptionType],
        addresses: typing.Optional[typing.List[HostAddress]],
        enc_authorization_data: typing.Optional[EncryptedData],
        additional_tickets: typing.Optional[typing.List[Ticket]],
    ) -> None:
        self.kdc_options = kdc_options
        self.cname = cname
        self.realm = realm
        self.sname = sname
        self.postdated_from = postdated_from
        self.postdated_till = postdated_till
        self.rtime = rtime
        self.nonce = nonce
        self.etype = etype
        self.addresses = addresses
        self.enc_authorization_data = enc_authorization_data
        self.additional_tickets = additional_tickets

    @staticmethod
    def unpack(value: typing.Union[ASN1Value, bytes]) -> "KdcReqBody":
        sequence = unpack_asn1_tagged_sequence(value)

        def unpack_kdc_options(value: typing.Union[ASN1Value, bytes]) -> int:
            b_data = unpack_asn1_bit_string(value)
            return struct.unpack(">I", b_data)[0]

        def unpack_etype(value: typing.Union[ASN1Value, bytes]) -> typing.List[KerberosEncryptionType]:
            return [KerberosEncryptionType(unpack_asn1_integer(e)) for e in unpack_asn1_sequence(value)]

        def unpack_addresses(value: typing.Union[ASN1Value, bytes]) -> typing.List[HostAddress]:
            return [unpack_hostname(h) for h in unpack_asn1_sequence(value)]

        def unpack_ticket(value: typing.Union[ASN1Value, bytes]) -> typing.List[Ticket]:
            return [Ticket.unpack(t) for t in unpack_asn1_sequence(value)]

        kdc_options = get_sequence_value(sequence, 0, "KDC-REQ-BODY", "kdc-options", unpack_kdc_options)
        cname = get_sequence_value(sequence, 1, "KDC-REQ-BODY", "cname", unpack_principal_name)
        realm = get_sequence_value(sequence, 2, "KDC-REQ-BODY", "realm", unpack_asn1_general_string)
        sname = get_sequence_value(sequence, 3, "KDC-REQ-BODY", "sname", unpack_principal_name)
        postdated_from = get_sequence_value(sequence, 4, "KDC-REQ-BODY", "from", unpack_asn1_generalized_time)
        postdated_till = get_sequence_value(sequence, 5, "KDC-REQ-BODY", "till", unpack_asn1_generalized_time)
        rtime = get_sequence_value(sequence, 6, "KDC-REQ-BODY", "rtime", unpack_asn1_generalized_time)
        nonce = get_sequence_value(sequence, 7, "KDC-REQ-BODY", "nonce", unpack_asn1_integer)
        etype = get_sequence_value(sequence, 8, "KDC-REQ-BODY", "etype", unpack_etype)
        addresses = get_sequence_value(sequence, 9, "KDC-REQ-BODY", "addresses", unpack_addresses)
        enc_auth_data = get_sequence_value(sequence, 10, "KDC-REQ-BODY", "enc-authorization-data", EncryptedData.unpack)
        additional_tickets = get_sequence_value(sequence, 11, "KDC-REQ-BODY", "additional-tickets", unpack_ticket)

        return KdcReqBody(
            kdc_options,
            cname,
            realm,
            sname,
            postdated_from,
            postdated_till,
            rtime,
            nonce,
            etype,
            addresses,
            enc_auth_data,
            additional_tickets,
        )



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import struct
import typing

from spnego._ntlm_raw.crypto import RC4Handle, crc32, hmac_md5, rc4
from spnego._ntlm_raw.messages import NegotiateFlags
from spnego.exceptions import OperationNotAvailableError


def seal(flags: int, handle: RC4Handle, signing_key: bytes, seq_num: int, b_data: bytes) -> typing.Tuple[bytes, bytes]:
    """Create a sealed NTLM message.

    Creates a sealed NTLM message as documented at `NTLM Message Confidentiality`_.

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        handle: The RC4 handle for the negotiated context.
        signing_key: The key used to sign the message.
        seq_num: The sequence number for the message.
        b_data: The data/message bytes to seal.

    Returns:
        Tuple[bytes, bytes]: The sealed message bytes and the message signature.

    .. _NTLM Message Confidentiality:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/115f9c7d-bc30-4262-ae96-254555c14ea6
    """
    seal_msg = rc4(handle, b_data)
    signature = sign(flags, handle, signing_key, seq_num, b_data)
    return seal_msg, signature


def sign(
    flags: int,
    handle: RC4Handle,
    signing_key: bytes,
    seq_num: int,
    b_data: bytes,
) -> bytes:
    """Create a NTLM signature.

    Creates a NTLM signature as documented at `NTLM Message Integrity`_ and appends it to the end of the message.

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        handle: The RC4 handle for the negotiated context.
        signing_key: The key used to sign the message.
        seq_num: The sequence number for the signature.
        b_data: The data/message bytes to sign.

    Returns:
        bytes: The data with the signature appended.

    .. _NTLM Message Integrity:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/131b0062-7958-460e-bca5-c7a9f9086652
    """
    if flags & NegotiateFlags.sign == 0:
        if flags & NegotiateFlags.always_sign == 0:
            raise OperationNotAvailableError(context_msg="Signing without integrity.")

        # This is the behaviour seen with SSPI when signing data with NTLMSSP_NEGOTIATE_ALWAYS_SIGN.
        return b"\x01" + b"\x00" * 15

    elif flags & NegotiateFlags.extended_session_security:
        return _mac_with_ess(flags, handle, signing_key, seq_num, b_data)

    else:
        return _mac_without_ess(handle, seq_num, b_data)


def _mac_without_ess(
    handle: RC4Handle,
    seq_num: int,
    b_data: bytes,
) -> bytes:
    """NTLM MAC without Extended Session Security

    Generates the NTLM signature when Extended Session Security has not been negotiated. The structure of the signature
    is documented at `NTLM signature without ESS`_.

    The algorithm as documented by `MAC without ESS`_ is::

        Define MAC(Handle, SigningKey, SeqNum, Message) as
            Set NTLMSSP_MESSAGE_SIGNATURE.Version to 0x00000001
            Set NTLMSSP_MESSAGE_SIGNATURE.Checksum to CRC32(Message)
            Set NTLMSSP_MESSAGE_SIGNATURE.RandomPad RC4(Handle, RandomPad)
            Set NTLMSSP_MESSAGE_SIGNATURE.Checksum to RC4(Handle, NTLMSSP_MESSAGE_SIGNATURE.Checksum)
            Set NTLMSSP_MESSAGE_SIGNATURE.SeqNum to RC4(Handle, 0x00000000)

            If (connection oriented)
                Set NTLMSSP_MESSAGE_SIGNATURE.SeqNum to NTLMSSP_MESSAGE_SIGNATURE.SeqNum XOR SeqNum
                Set SeqNum to SeqNum + 1

            Else
                Set NTLMSSP_MESSAGE_SIGNATURE.SeqNum to NTLMSSP_MESSAGE_SIGNATURE.SeqNum XOR (app supplied SeqNum)

            Endif

            Set NTLMSSP_MESSAGE_SIGNATURE.RandomPad to 0

        EndDefine

    Args:
        handle: The RC4 handle for the negotiated context.
        seq_num: The sequence number for the signature.
        b_data: The data/message bytes to sign.

    Returns:
        bytes: The NTLM signature.

    .. _NTLM signature without ESS:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83fbd0e7-8ab0-4873-8cbe-795249b46b8a

    .. _MAC without ESS:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/0b1fb6a6-7224-4d5b-af35-fdd45c0791e5
    """
    checksum = crc32(b_data)
    rc4(handle, b"\x00\x00\x00\x00")  # NTLMSSP_MESSAGE_SIGNATURE.RandomPad RC4(Handle, RandomPad)
    checksum = rc4(handle, checksum)

    temp_seq_num = struct.unpack("<I", rc4(handle, b"\x00\x00\x00\x00"))[0]
    b_seq_num = struct.pack("<I", temp_seq_num ^ seq_num)

    return b"\x01\x00\x00\x00" + b"\x00\x00\x00\x00" + checksum + b_seq_num


def _mac_with_ess(flags: int, handle: RC4Handle, signing_key: bytes, seq_num: int, b_data: bytes) -> bytes:
    """NTLM MAC with Extended Session Security

    Generates the NTLM signature when Extended Session Security has been negotiated. The structure of the signature is
    documented at `NTLM signature with ESS`_.

    The algorithm as documented by `MAC with ESS`_ is::

        Define MAC(Handle, SigningKey, SeqNum, Message) as
            Set NTLMSSP_MESSAGE_SIGNATURE.Version to 0x00000001
            Set NTLMSSP_MESSAGE_SIGNATURE.Checksum to HMAC_MD5(SigningKey, ConcatenationOf(SeqNum, Message))[0..7]
            Set NTLMSSP_MESSAGE_SIGNATURE.SeqNum to SeqNum
            Set SeqNum to SeqNum + 1
        EndDefine

        # When NegotiateFlags.key_exch

        Define MAC(Handle, SigningKey, SeqNum, Message) as
            Set NTLMSSP_MESSAGE_SIGNATURE.Version to 0x00000001
            Set NTLMSSP_MESSAGE_SIGNATURE.Checksum to RC4(Handle,
                HMAC_MD5(SigningKey, ConcatenationOf(SeqNum, Message))[0..7])
            Set NTLMSSP_MESSAGE_SIGNATURE.SeqNum to SeqNum
            Set SeqNum to SeqNum + 1
        EndDefine

    Args:
        flags: The negotiated flags between the initiator and acceptor.
        handle: The RC4 handle for the negotiated context.
        signing_key: The key used to sign the message.
        seq_num: The sequence number for the signature.
        b_data: The data/message bytes to sign.

    Returns:
        bytes: The NTLM with ESS signature.

    .. _NTLM signature with ESS:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/2c3b4689-d6f1-4dc6-85c9-0bf01ea34d9f

    .. _MAC with ESS:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/a92716d5-d164-4960-9e15-300f4eef44a8
    """
    b_seq_num = struct.pack("<I", seq_num)

    checksum = hmac_md5(signing_key, b_seq_num + b_data)[:8]
    if flags & NegotiateFlags.key_exch:
        checksum = handle.update(checksum)

    return b"\x01\x00\x00\x00" + checksum + b_seq_num



# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Credentials that can be used by a client.

These are the various credentials that can be used as a client to authenticate
with a service. See the docstring for each credential type to identify what
credentials are usable for each authenticaton protocol.
"""

import dataclasses
import typing

from spnego._ntlm_raw.crypto import is_ntlm_hash
from spnego.exceptions import InvalidCredentialError, NoCredentialError

# FUTURE: Add Anonymous credential class


@dataclasses.dataclass
class Password:
    """Plaintext username and password credential.

    This is the simple username and password credential that is supported by
    all authentication protocols.

    Attributes:
        username: The username.
        password: The password for the user.
    """

    username: str
    password: str = dataclasses.field(repr=False)

    @property
    def supported_protocols(self) -> typing.List[str]:
        """List of protocols the credential can be used for."""
        return ["credssp", "kerberos", "ntlm"]


@dataclasses.dataclass
class CredentialCache:
    """Cached credential.

    Uses the provider specific cached credential. Can also specify a username
    to select a specific cached credential.

    Attributes:
        username: Optional username used to select a specific credential in the
            cache.
    """

    username: typing.Optional[str] = None

    @property
    def supported_protocols(self) -> typing.List[str]:
        return ["kerberos", "ntlm"]


@dataclasses.dataclass
class NTLMHash:
    """NTLM LM/NT Hash credential.

    Used with :class:`NTLMProxy` for NTLM authentication backed by an NT and/or
    LM hash value. In modern iterations of NTLM only the NT hash needs to be
    specified and LM is ignored but both can be specified and the NTLM code
    will use them as necessary.

    Attributes:
        username: The username the hashes are for.
        lm_hash: The LM hash as a hex string, can be `None` in most cases.
        nt_hash: The NT hash as a hex string.
    """

    username: str
    lm_hash: typing.Optional[str] = dataclasses.field(default=None, repr=False)
    nt_hash: typing.Optional[str] = dataclasses.field(default=None, repr=False)

    @property
    def supported_protocols(self) -> typing.List[str]:
        """List of protocols the credential can be used for."""
        return ["ntlm"]


@dataclasses.dataclass
class KerberosKeytab:
    """Kerberos Keytab Credential.

    Used with :class:`GSSAPIProxy` for Kerberos authentication. It is used to
    retrieve a Kerberos ticket using a keytab for authentication rather than a
    password. The keytab value is specified in the form ``TYPE:RESIDUAL`` where
    the ``TYPE`` supported is down to the installed Kerberos/GSSAPI
    implementation and ``RESIDUAL`` is a value specific to the type. Common
    types are:

        FILE: The value is the path to a keytab.
        MEMORY: The value is a unique identifier to a keytab stored in memory
            of the current process. It must be resolvable by the linked GSSAPI
            provider that this library uses.

    There are other ccache types but they are mostly platform or GSSAPI
    implementation specific.

    .. Note:
        This only works on Linux, Windows does not have the concept of a
        keytab.

    Attributes:
        keytab: The keytab to use for authentication. The path will not be
            expanded of have variables substituted so should be the absolute
            path to the keytab.
        principal: The Kerberos principal to get the credential for. Should be
            in the UPN form `username@REALM.COM`. Set to `None` to use the
            first keytab entry.
    """

    keytab: str
    principal: typing.Optional[str] = None

    @property
    def supported_protocols(self) -> typing.List[str]:
        """List of protocols the credential can be used for."""
        return ["kerberos"]


@dataclasses.dataclass
class KerberosCCache:
    """Kerberos CCache Credential.

    Used with :class:`GSSAPIProxy` for Kerberos authentication. It is used to
    specify the credential cache that has the stored Kerberos credential for
    authentication. The ccache value is specified in the form ``TYPE:RESIDUAL``
    where the ``TYPE`` supported is down to the installed Kerberos/GSSAPI
    implementation and ``RESIDUAL`` is a value specific to the type. Common
    types are:

        DIR: The value is the path to a directory containing a collection of
            `FILE` caches.
        FILE: The value is the path to an individual cache.
        MEMORY: The value is a unique identifier to a cache stored in memory of
            the current process. It must be resolvable by the linked GSSAPI
            provider that this library uses.

    There are other ccache types but they are mostly platform or GSSAPI
    implementation specific.

    .. Note:
        This only works on Linux, Windows does not have the concept of
        separate CCaches.

    Attributes:
        ccache: The ccache in the form ``TYPE:RESIDUAL`` to use for a Kerberos
            credential. The path will not be expanded of have variables
            substituted so should be the absolute path to the ccache.
        principal: Optional principal to get in the credential cache specified.
    """

    ccache: str
    principal: typing.Optional[str] = None

    @property
    def supported_protocols(self) -> typing.List[str]:
        """List of protocols the credential can be used for."""
        return ["kerberos"]


# FUTURE: Expose this as an official credential
# @dataclasses.dataclass
# class CredSSPSmartCard:
#     """CredSSP Smart Card Credential.

#     Used with :class:`CredSSPProxy` for CredSSP authentication. It is used to
#     delegate a smart card credential to the peer rather than a
#     username/password. It must be supplied with another NTLM/Kerberos supported
#     credential as this is only used for the delegation part of the CredSSP
#     authentication process.

#     The values are modeled after `MS-CSSP TSSmartCardCreds`_.

#     Attributes:
#         pin: The user's smart card PIN.
#         key_spec: The specification of the user's smart card.
#         card_name: The name of the smart card.
#         reader_name: The name of the smart card reader.
#         container_name: Name of the certificate container.
#         csp_name: Name of the CSP.
#         user_hint: The user's account hint.
#         domain_hint: The user's domain name hint.

#     .. _MS-CSSP TSSmartCardCreds:
#         https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/4251d165-cf01-4513-a5d8-39ee4a98b7a4
#     """
#     pin: bytes
#     key_spec: int
#     card_name: typing.Optional[bytes] = None
#     reader_name: typing.Optional[bytes] = None
#     container_name: typing.Optional[bytes] = None
#     csp_name: typing.Optional[bytes] = None
#     user_hint: typing.Optional[bytes] = None
#     domain_hint: typing.Optional[bytes] = None

#     @property
#     def supported_protocols(self) -> typing.List[str]:
#         return ["credssp"]


# @dataclasses.dataclass
# class CredSSPRemoteGuardPackage:
#     """CredSSP Remote Guard Credential.

#     Used with :class:`CredSSPProxy` for CredSSP authentication. It is used to
#     delegate credential protected by `Remote Credential Guard`_. It must be
#     supplied with another NTLM/Kerberos supported credential as this is only
#     used for the delegation part of the CredSSP authentication process.
#     Multiple instance of this can be passed in with the first entry being the
#     logon credential and the remaining instances being the supplemental
#     credentials.

#     The values are modeled after `MS-CSSP TSRemoteGuardPackageCred`_.

#     .. _Remote Credential Guard:
#         https://docs.microsoft.com/en-us/windows/security/identity-protection/remote-credential-guard

#     .. _MS-CSSP TSRemoteGuardPackageCred:
#         https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/173eee44-1a2c-463f-b909-c15db01e68d7
#     """
#     package_name: str
#     cred_buffer: bytes

#     @property
#     def supported_protocols(self) -> typing.List[str]:
#         return ["credssp"]


Credential = typing.Union[CredentialCache, KerberosCCache, KerberosKeytab, NTLMHash, Password]


def unify_credentials(
    username: typing.Optional[typing.Union[str, Credential, typing.List[Credential]]] = None,
    password: typing.Optional[str] = None,
    required_protocol: typing.Optional[str] = None,
) -> typing.List[Credential]:
    """Process user input credentials.

    Converts the user facing credential input into a list of credentials that
    is known by spnego. Also filters out any duplicate credentials/ones that
    are rendered obsolete by any ones preceding it. For example NTLMHash won't
    be used if it's specified after Password in the credential list.

    Args:
        username: The username or list of credentials to process.
        password: The password for username when it's a string.
        required_protocol: Optionally checks that at least 1 credential must
            support the protocol specified.

    Returns:
        typing.List[Credential]: A list of credentials based on the free-form
        input.
    """
    if username:
        if isinstance(username, str):
            if not password:
                username = [CredentialCache(username=username)]
            elif is_ntlm_hash(password):
                lm, nt = password.split(":", 1)
                username = [NTLMHash(username=username, lm_hash=lm, nt_hash=nt)]
            else:
                username = [Password(username=username, password=password)]

        elif not isinstance(username, list):
            username = [username]

    else:
        username = [CredentialCache()]

    credentials: typing.List[Credential] = []
    used_protocols: typing.Set[str] = set()
    for cred in username:
        if not isinstance(cred, (CredentialCache, KerberosCCache, KerberosKeytab, NTLMHash, Password)):
            raise InvalidCredentialError(
                context_msg="Invalid username/credential specified, must be a string or Credential object."
            )

        # Remove any credential that only implements protocols that are already covered by one before it.
        # FUTURE: This might be problematic with the CredSSP ones
        cred_useful = False
        for cred_protocol in cred.supported_protocols:
            if cred_protocol not in used_protocols:
                used_protocols.add(cred_protocol)
                cred_useful = True

        if cred_useful:
            credentials.append(cred)

    if required_protocol and required_protocol not in used_protocols:
        found_protocols = ", ".join(sorted(used_protocols))
        raise NoCredentialError(
            context_msg=f"A credential for {required_protocol} is needed but only found "
            f"credentials for {found_protocols}"
        )

    return credentials



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import copy
import logging
import sys
import typing

from spnego._context import (
    IOV,
    ContextProxy,
    ContextReq,
    GSSMech,
    IOVUnwrapResult,
    IOVWrapResult,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
    wrap_system_error,
)
from spnego._credential import (
    Credential,
    CredentialCache,
    KerberosCCache,
    KerberosKeytab,
    Password,
    unify_credentials,
)
from spnego._text import to_bytes, to_text
from spnego.channel_bindings import GssChannelBindings
from spnego.exceptions import GSSError as NativeError
from spnego.exceptions import (
    InvalidCredentialError,
    NegotiateOptions,
    NoContextError,
    SpnegoError,
)
from spnego.iov import BufferType, IOVBuffer, IOVResBuffer

log = logging.getLogger(__name__)

HAS_GSSAPI = True
GSSAPI_IMP_ERR = None
try:
    import gssapi
    import krb5
    from gssapi.raw import ChannelBindings, GSSError
    from gssapi.raw import exceptions as gss_errors
    from gssapi.raw import inquire_sec_context_by_oid, set_cred_option
except ImportError as e:
    GSSAPI_IMP_ERR = str(e)
    HAS_GSSAPI = False
    log.debug("Python gssapi not available, cannot use any GSSAPIProxy protocols: %s" % e)


HAS_IOV = True
GSSAPI_IOV_IMP_ERR = None
try:
    from gssapi.raw import IOV as GSSIOV
    from gssapi.raw import IOVBuffer as GSSIOVBuffer
    from gssapi.raw import IOVBufferType, unwrap_iov, wrap_iov
except ImportError as err:
    GSSAPI_IOV_IMP_ERR = sys.exc_info()
    HAS_IOV = False
    log.debug("Python gssapi IOV extension not available: %s" % str(GSSAPI_IOV_IMP_ERR[1]))

_GSS_C_INQ_SSPI_SESSION_KEY = "1.2.840.113554.1.2.2.5.5"

_GSS_KRB5_CRED_NO_CI_FLAGS_X = "1.2.752.43.13.29"


def _create_iov_result(iov: "GSSIOV") -> typing.Tuple[IOVResBuffer, ...]:
    """Converts GSSAPI IOV buffer to generic IOVBuffer result."""
    buffers = []
    for i in iov:
        buffer_entry = IOVResBuffer(type=BufferType(i.type), data=i.value)
        buffers.append(buffer_entry)

    return tuple(buffers)


def _get_gssapi_credential(
    mech: "gssapi.OID",
    usage: str,
    credentials: typing.List[Credential],
    context_req: typing.Optional[ContextReq] = None,
) -> typing.Optional["gssapi.creds.Credentials"]:
    """Gets the GSSAPI credential.

    Will get a GSSAPI credential for the mech specified. If the username and password is specified then a new
    set of credentials are explicitly required for the mech specified. Otherwise the credentials are retrieved based on
    the credential type specified.

    Args:
        mech: The mech OID to get the credentials for, only Kerberos is supported.
        usage: Either `initiate` for a client context or `accept` for a server context.
        credentials: List of credentials to retreive from.
        context_req: Context requirement flags that can control how the credential is retrieved.

    Returns:
        gssapi.creds.Credentials: The credential set that was created/retrieved.
    """
    name_type = getattr(gssapi.NameType, "user" if usage == "initiate" else "hostbased_service")
    forwardable = bool(context_req and (context_req & ContextReq.delegate or context_req & ContextReq.delegate_policy))

    for cred in credentials:
        if isinstance(cred, CredentialCache):
            principal = None
            if cred.username:
                principal = gssapi.Name(base=cred.username, name_type=name_type)

            elif usage == "initiate":
                # https://github.com/jborean93/pyspnego/issues/15
                # Using None as a credential when creating the sec context is better than getting the default
                # credential as the former takes into account the target SPN when selecting the principal to use.
                return None

            gss_cred = gssapi.Credentials(name=principal, usage=usage, mechs=[mech])

            # We don't need to check the actual lifetime, just trying to get the valid will have gssapi check the
            # lifetime and raise an ExpiredCredentialsError if it is expired.
            _ = gss_cred.lifetime

            return gss_cred

        elif isinstance(cred, KerberosCCache):
            if usage != "initiate":
                log.debug("Skipping %s as it can only be used for an initiate Kerberos context", cred)
                continue

            ctx = krb5.init_context()
            ccache = krb5.cc_resolve(ctx, to_bytes(cred.ccache))
            krb5_principal: typing.Optional[krb5.Principal] = None
            if cred.principal:
                krb5_principal = krb5.parse_name_flags(ctx, to_bytes(cred.principal))

            return gssapi.Credentials(base=_gss_acquire_cred_from_ccache(ccache, krb5_principal), usage=usage)

        elif isinstance(cred, (KerberosKeytab, Password)):
            if usage != "initiate":
                log.debug("Skipping %s as it can only be used for an initiate Kerberos context", cred)
                continue

            if isinstance(cred, KerberosKeytab):
                username = cred.principal or ""
                password = cred.keytab
                is_keytab = True
            else:
                username = cred.username
                password = cred.password
                is_keytab = False

            raw_cred = _kinit(
                to_bytes(username),
                to_bytes(password),
                forwardable=forwardable,
                is_keytab=is_keytab,
            )

            return gssapi.Credentials(base=raw_cred, usage=usage)

        else:
            log.debug("Skipping credential %s as it does not support required mech type", cred)
            continue

    raise InvalidCredentialError(context_msg="No applicable credentials available")


def _gss_sasl_description(mech: "gssapi.OID") -> typing.Optional[bytes]:
    """Attempts to get the SASL description of the mech specified."""
    try:
        res = _gss_sasl_description.result  # type: ignore
        return res[mech.dotted_form]

    except (AttributeError, KeyError):
        res = getattr(_gss_sasl_description, "result", {})

    try:
        sasl_desc = gssapi.raw.inquire_saslname_for_mech(mech).mech_description
    except Exception as e:
        log.debug("gss_inquire_saslname_for_mech(%s) failed: %s" % (mech.dotted_form, str(e)))
        sasl_desc = None

    res[mech.dotted_form] = sasl_desc
    _gss_sasl_description.result = res  # type: ignore
    return _gss_sasl_description(mech)


def _kinit(
    username: bytes,
    password: bytes,
    forwardable: typing.Optional[bool] = None,
    is_keytab: bool = False,
) -> "gssapi.raw.Creds":
    """Gets a Kerberos credential.

    This will get the GSSAPI credential that contains the Kerberos TGT inside
    it. This is used instead of gss_acquire_cred_with_password as the latter
    does not expose a way to request a forwardable ticket or to retrieve a TGT
    from a keytab. This way makes it possible to request whatever is needed
    before making it usable in GSSAPI.

    Args:
        username: The username to get the credential for.
        password: The password to use to retrieve the credential.
        forwardable: Whether to request a forwardable credential.
        is_keytab: Whether password is a keytab or just a password.

    Returns:
        gssapi.raw.Creds: The GSSAPI credential for the Kerberos mech.
    """
    ctx = krb5.init_context()

    kt: typing.Optional[krb5.KeyTab] = None
    princ: typing.Optional[krb5.Principal] = None
    if is_keytab:
        kt = krb5.kt_resolve(ctx, password)

        # If the username was not specified get the principal of the first entry.
        if not username:
            # The principal handle is deleted once the entry is deallocated. Make sure it is stored in a var before
            # being copied.
            first_entry = list(kt)[0]
            princ = copy.copy(first_entry.principal)

    if not princ:
        princ = krb5.parse_name_flags(ctx, username)

    init_opt = krb5.get_init_creds_opt_alloc(ctx)

    if hasattr(krb5, "get_init_creds_opt_set_default_flags"):
        # Heimdal requires this to be set in order to load the default options from krb5.conf. This follows the same
        # code that it's own gss_acquire_cred_with_password does.
        realm = krb5.principal_get_realm(ctx, princ)
        krb5.get_init_creds_opt_set_default_flags(ctx, init_opt, b"gss_krb5", realm)

    krb5.get_init_creds_opt_set_canonicalize(init_opt, True)
    if forwardable is not None:
        krb5.get_init_creds_opt_set_forwardable(init_opt, forwardable)

    if kt:
        cred = krb5.get_init_creds_keytab(ctx, princ, init_opt, keytab=kt)
    else:
        cred = krb5.get_init_creds_password(ctx, princ, init_opt, password=password)

    mem_ccache = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, mem_ccache, princ)
    krb5.cc_store_cred(ctx, mem_ccache, cred)

    return _gss_acquire_cred_from_ccache(mem_ccache, None)


def _gss_acquire_cred_from_ccache(
    ccache: "krb5.CCache",
    principal: typing.Optional["krb5.Principal"],
) -> "gssapi.raw.Creds":
    """Acquire GSSAPI credential from CCache.

    Args:
        ccache: The CCache to acquire the credential from.
        principal: The optional principal to acquire the cred for.

    Returns:
        gssapi.raw.Creds: The GSSAPI credentials from the ccache.
    """
    # acquire_cred_from is less dangerous than krb5_import_cred which uses a raw pointer to access the ccache. Heimdal
    # has only recently added this API (not in a release as of 2021) so there's a fallback to the latter API.
    if hasattr(gssapi.raw, "acquire_cred_from"):
        kerberos = gssapi.OID.from_int_seq(GSSMech.kerberos.value)
        name = None
        if principal:
            name = gssapi.Name(base=to_text(principal.name), name_type=gssapi.NameType.user)

        ccache_name = ccache.name or b""
        if ccache.cache_type:
            ccache_name = ccache.cache_type + b":" + ccache_name

        return gssapi.raw.acquire_cred_from(
            {b"ccache": ccache_name},
            name=name,
            mechs=[kerberos],
            usage="initiate",
        ).creds

    else:
        gssapi_creds = gssapi.raw.Creds()
        gssapi.raw.krb5_import_cred(
            gssapi_creds, cache=ccache.addr, keytab_principal=principal.addr if principal else None
        )

        return gssapi_creds


class GSSAPIProxy(ContextProxy):
    """GSSAPI proxy class for GSSAPI on Linux.

    This proxy class for GSSAPI exposes GSSAPI calls into a common interface for Kerberos authentication. This context
    uses the Python gssapi library to interface with the gss_* calls to provider Kerberos.
    """

    def __init__(
        self,
        username: typing.Optional[typing.Union[str, Credential, typing.List[Credential]]] = None,
        password: typing.Optional[str] = None,
        hostname: typing.Optional[str] = None,
        service: typing.Optional[str] = None,
        channel_bindings: typing.Optional[GssChannelBindings] = None,
        context_req: ContextReq = ContextReq.default,
        usage: str = "initiate",
        protocol: str = "kerberos",
        options: NegotiateOptions = NegotiateOptions.none,
        **kwargs: typing.Any,
    ) -> None:

        if not HAS_GSSAPI:
            raise ImportError("GSSAPIProxy requires the Python gssapi library: %s" % GSSAPI_IMP_ERR)

        credentials = unify_credentials(username, password)
        super(GSSAPIProxy, self).__init__(
            credentials, hostname, service, channel_bindings, context_req, usage, protocol, options
        )

        self._mech = gssapi.OID.from_int_seq(GSSMech.kerberos.value)

        gssapi_credential = kwargs.get("_gssapi_credential", None)
        if not gssapi_credential:
            try:
                gssapi_credential = _get_gssapi_credential(
                    self._mech,
                    self.usage,
                    credentials=credentials,
                    context_req=context_req,
                )
            except GSSError as gss_err:
                raise SpnegoError(base_error=gss_err, context_msg="Getting GSSAPI credential") from gss_err

        if context_req & ContextReq.no_integrity and self.usage == "initiate":
            if gssapi_credential is None:
                gssapi_credential = gssapi.Credentials(usage=self.usage, mechs=[self._mech])

            set_cred_option(
                gssapi.OID.from_int_seq(_GSS_KRB5_CRED_NO_CI_FLAGS_X),
                gssapi_credential,
            )

        self._credential = gssapi_credential
        self._context: typing.Optional[gssapi.SecurityContext] = None

    @classmethod
    def available_protocols(cls, options: typing.Optional[NegotiateOptions] = None) -> typing.List[str]:
        # We can't offer Kerberos if the caller requires WinRM wrapping and IOV isn't available.
        avail = []
        if not (options and options & NegotiateOptions.wrapping_winrm and not HAS_IOV):
            avail.append("kerberos")

        return avail

    @classmethod
    def iov_available(cls) -> bool:
        return HAS_IOV

    @property
    def client_principal(self) -> typing.Optional[str]:
        # Looks like a bug in python-gssapi where the value still has the terminating null char.
        if self._context and self.usage == "accept":
            return to_text(self._context.initiator_name).rstrip("\x00")
        else:
            return None

    @property
    def complete(self) -> bool:
        return self._context is not None and self._context.complete

    @property
    def negotiated_protocol(self) -> typing.Optional[str]:
        return "kerberos"

    @property
    @wrap_system_error(NativeError, "Retrieving session key")
    def session_key(self) -> bytes:
        if self._context:
            return inquire_sec_context_by_oid(self._context, gssapi.OID.from_int_seq(_GSS_C_INQ_SSPI_SESSION_KEY))[0]
        else:
            raise NoContextError(context_msg="Retrieving session key failed as no context was initialized")

    def new_context(self) -> "GSSAPIProxy":
        return GSSAPIProxy(
            hostname=self._hostname,
            service=self._service,
            channel_bindings=self.channel_bindings,
            context_req=self.context_req,
            usage=self.usage,
            protocol=self.protocol,
            options=self.options,
            _gssapi_credential=self._credential,
        )

    @wrap_system_error(NativeError, "Processing security token")
    def step(
        self,
        in_token: typing.Optional[bytes] = None,
        *,
        channel_bindings: typing.Optional[GssChannelBindings] = None,
    ) -> typing.Optional[bytes]:
        if not self._is_wrapped:
            log.debug("GSSAPI step input: %s", base64.b64encode(in_token or b"").decode())

        if not self._context:
            context_kwargs: typing.Dict[str, typing.Any] = {}

            channel_bindings = channel_bindings or self.channel_bindings
            if channel_bindings:
                context_kwargs["channel_bindings"] = ChannelBindings(
                    initiator_address_type=channel_bindings.initiator_addrtype,
                    initiator_address=channel_bindings.initiator_address,
                    acceptor_address_type=channel_bindings.acceptor_addrtype,
                    acceptor_address=channel_bindings.acceptor_address,
                    application_data=channel_bindings.application_data,
                )

            if self.usage == "initiate":
                spn = "%s@%s" % (self._service or "host", self._hostname or "unspecified")
                context_kwargs["name"] = gssapi.Name(spn, name_type=gssapi.NameType.hostbased_service)
                context_kwargs["mech"] = self._mech
                context_kwargs["flags"] = self._context_req

            self._context = gssapi.SecurityContext(creds=self._credential, usage=self.usage, **context_kwargs)

        out_token = self._context.step(in_token)

        try:
            self._context_attr = int(self._context.actual_flags)
        except gss_errors.MissingContextError:  # pragma: no cover
            # MIT krb5 before 1.14.x will raise this error if the context isn't
            # complete. We should only treat it as an error if it happens when
            # the context is complete (last step).
            # https://github.com/jborean93/pyspnego/issues/55
            if self._context.complete:
                raise

        if not self._is_wrapped:
            log.debug("GSSAPI step output: %s", base64.b64encode(out_token or b"").decode())

        return out_token

    @wrap_system_error(NativeError, "Wrapping data")
    def wrap(self, data: bytes, encrypt: bool = True, qop: typing.Optional[int] = None) -> WrapResult:
        if not self._context:
            raise NoContextError(context_msg="Cannot wrap until context has been established")
        res = gssapi.raw.wrap(self._context, data, confidential=encrypt, qop=qop)

        return WrapResult(data=res.message, encrypted=res.encrypted)

    @wrap_system_error(NativeError, "Wrapping IOV buffer")
    def wrap_iov(
        self,
        iov: typing.Iterable[IOV],
        encrypt: bool = True,
        qop: typing.Optional[int] = None,
    ) -> IOVWrapResult:
        if not self._context:
            raise NoContextError(context_msg="Cannot wrap until context has been established")

        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = GSSIOV(*buffers, std_layout=False)
        encrypted = wrap_iov(self._context, iov_buffer, confidential=encrypt, qop=qop)

        return IOVWrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypted)

    def wrap_winrm(self, data: bytes) -> WinRMWrapResult:
        iov = self.wrap_iov([BufferType.header, data, BufferType.padding]).buffers
        header = iov[0].data or b""
        enc_data = iov[1].data or b""
        padding = iov[2].data or b""

        return WinRMWrapResult(header=header, data=enc_data + padding, padding_length=len(padding))

    @wrap_system_error(NativeError, "Unwrapping data")
    def unwrap(self, data: bytes) -> UnwrapResult:
        if not self._context:
            raise NoContextError(context_msg="Cannot unwrap until context has been established")

        res = gssapi.raw.unwrap(self._context, data)

        return UnwrapResult(data=res.message, encrypted=res.encrypted, qop=res.qop)

    @wrap_system_error(NativeError, "Unwrapping IOV buffer")
    def unwrap_iov(
        self,
        iov: typing.Iterable[IOV],
    ) -> IOVUnwrapResult:
        if not self._context:
            raise NoContextError(context_msg="Cannot unwrap until context has been established")

        buffers = self._build_iov_list(iov, self._convert_iov_buffer)
        iov_buffer = GSSIOV(*buffers, std_layout=False)
        res = unwrap_iov(self._context, iov_buffer)

        return IOVUnwrapResult(buffers=_create_iov_result(iov_buffer), encrypted=res.encrypted, qop=res.qop)

    def unwrap_winrm(self, header: bytes, data: bytes) -> bytes:
        # This is an extremely weird setup, Kerberos depends on the underlying provider that is used. Right now the
        # proper IOV buffers required to work on both AES and RC4 encrypted only works for MIT KRB5 whereas Heimdal
        # fails. It currently mandates a padding buffer of a variable size which we cannot achieve in the way that
        # WinRM encrypts the data. This is fixed in the source code but until it is widely distributed we just need to
        # use a way that is known to just work with AES. To ensure that MIT works on both RC4 and AES we check the
        # description which differs between the 2 implemtations. It's not perfect but I don't know of another way to
        # achieve this until more time has passed.
        # https://github.com/heimdal/heimdal/issues/739
        if not self._context:
            raise NoContextError(context_msg="Cannot unwrap until context has been established")

        sasl_desc = _gss_sasl_description(self._context.mech)

        # https://github.com/krb5/krb5/blob/f2e28f13156785851819fc74cae52100e0521690/src/lib/gssapi/krb5/gssapi_krb5.c#L686
        if sasl_desc and sasl_desc == b"Kerberos 5 GSS-API Mechanism":
            iov = self.unwrap_iov([(IOVBufferType.header, header), data, IOVBufferType.data]).buffers
            return iov[1].data or b""

        else:
            return self.unwrap(header + data).data

    @wrap_system_error(NativeError, "Signing message")
    def sign(self, data: bytes, qop: typing.Optional[int] = None) -> bytes:
        if not self._context:
            raise NoContextError(context_msg="Cannot sign until context has been established")

        return gssapi.raw.get_mic(self._context, data, qop=qop)

    @wrap_system_error(NativeError, "Verifying message")
    def verify(self, data: bytes, mic: bytes) -> int:
        if not self._context:
            raise NoContextError(context_msg="Cannot verify until context has been established")

        return gssapi.raw.verify_mic(self._context, data, mic)

    @property
    def _context_attr_map(self) -> typing.List[typing.Tuple[ContextReq, int]]:
        attr_map = [
            (ContextReq.delegate, "delegate_to_peer"),
            (ContextReq.mutual_auth, "mutual_authentication"),
            (ContextReq.replay_detect, "replay_detection"),
            (ContextReq.sequence_detect, "out_of_sequence_detection"),
            (ContextReq.confidentiality, "confidentiality"),
            (ContextReq.integrity, "integrity"),
            # Only present when the DCE extensions are installed.
            (ContextReq.identify, "identify"),
            # Only present with newer versions of python-gssapi https://github.com/pythongssapi/python-gssapi/pull/218.
            (ContextReq.delegate_policy, "ok_as_delegate"),
        ]
        attrs = []
        for spnego_flag, gssapi_name in attr_map:
            if hasattr(gssapi.RequirementFlag, gssapi_name):
                attrs.append((spnego_flag, getattr(gssapi.RequirementFlag, gssapi_name)))

        return attrs

    def _convert_iov_buffer(self, buffer: IOVBuffer) -> "GSSIOVBuffer":
        buffer_data = None
        buffer_alloc = False

        if isinstance(buffer.data, bytes):
            buffer_data = buffer.data
        elif isinstance(buffer.data, bool):
            buffer_alloc = buffer.data
        elif isinstance(buffer.data, int):
            # This shouldn't really occur on GSSAPI but is here to mirror what SSPI does.
            buffer_data = b"\x00" * buffer.data
        else:
            auto_alloc = [BufferType.header, BufferType.padding, BufferType.trailer]
            buffer_alloc = buffer.type in auto_alloc

        return GSSIOVBuffer(IOVBufferType(buffer.type), buffer_alloc, buffer_data)
